from dataclasses import dataclass
from typing import Optional, Tuple, List, Dict
import os

from AgentCrew.modules.agents.local_agent import LocalAgent
from AgentCrew.modules.chat.file_handler import FileHandler
from AgentCrew.modules.llm.model_registry import ModelRegistry
from AgentCrew.modules.llm.service_manager import ServiceManager
from AgentCrew.modules.chat.consolidation import ConversationConsolidator
from AgentCrew.modules.config import ConfigManagement
from AgentCrew.modules.mcpclient import MCPService
import shlex


@dataclass
class CommandResult:
    """Result of command processing."""

    handled: bool = False
    exit_flag: bool = False
    clear_flag: bool = False
    message: str = ""


class CommandProcessor:
    """Handles command processing for the message handler."""

    def __init__(self, message_handler):
        from AgentCrew.modules.chat.message import MessageHandler

        if isinstance(message_handler, MessageHandler):
            self.message_handler = message_handler

    async def process_command(self, user_input: str) -> CommandResult:
        """Process a command and return the result."""
        if self._is_exit_command(user_input):
            self.message_handler._notify("exit_requested")
            return CommandResult(handled=True, exit_flag=True)
        elif user_input.lower() == "/clear":
            self.message_handler.start_new_conversation()
            return CommandResult(handled=True, clear_flag=True)
        elif user_input.lower() == "/copy":
            self.message_handler._notify(
                "copy_requested", self.message_handler.latest_assistant_response
            )
            return CommandResult(handled=True, clear_flag=True)
        elif user_input.lower() == "/debug":
            self.message_handler._notify(
                "debug_requested", self.message_handler.agent.history
            )
            self.message_handler._notify(
                "debug_requested", self.message_handler.streamline_messages
            )
            return CommandResult(handled=True, clear_flag=True)
        elif user_input.lower().startswith("/think "):
            try:
                budget = user_input[7:].strip()
                self.message_handler.agent.configure_think(budget)
                self.message_handler._notify("think_budget_set", budget)
            except ValueError:
                self.message_handler._notify(
                    "error", "Invalid budget value. Please provide a number."
                )
            return CommandResult(handled=True, clear_flag=True)
        elif user_input.lower().startswith("/consolidate"):
            return await self._handle_consolidate_command(user_input)
        elif user_input.lower().startswith("/unconsolidate"):
            return await self._handle_unconsolidate_command(user_input)
        elif user_input.lower().startswith("/jump "):
            jumped = self._handle_jump_command(user_input)
            return CommandResult(handled=jumped, clear_flag=True)
        elif user_input.lower().startswith("/agent"):
            success, message = self._handle_agent_command(user_input)
            self.message_handler._notify(
                "agent_command_result", {"success": success, "message": message}
            )
            return CommandResult(handled=True, clear_flag=True)
        elif user_input.lower().startswith("/model"):
            exit_flag, clear_flag = self._handle_model_command(user_input)
            return CommandResult(
                handled=True, exit_flag=exit_flag, clear_flag=clear_flag
            )
        elif user_input.lower().startswith("/mcp"):
            exit_flag, clear_flag = await self._handle_mcp_command(user_input)
            return CommandResult(
                handled=True, exit_flag=exit_flag, clear_flag=clear_flag
            )
        elif user_input.startswith("/file "):
            return self._handle_file_command(user_input)
        elif user_input.startswith("/drop "):
            return self._handle_drop_command(user_input)
        elif user_input.lower() == "/voice":
            return await self._handle_voice_command(user_input)
        elif user_input.lower() == "/end_voice":
            return await self._handle_end_voice_command(user_input)
        return CommandResult(handled=False)

    def _is_exit_command(self, user_input: str) -> bool:
        return user_input.lower() in ["/exit", "/quit"]

    async def _handle_consolidate_command(self, user_input: str) -> CommandResult:
        """Handle consolidate command."""
        try:
            parts = user_input.split()
            if len(parts) == 1:
                preserve_count = 10
            else:
                preserve_count = int(parts[1])

            if isinstance(self.message_handler.agent, LocalAgent):
                consolidator = ConversationConsolidator(self.message_handler.agent.llm)

                result = await consolidator.consolidate(
                    self.message_handler.streamline_messages, preserve_count
                )

                if result["success"]:
                    self.message_handler.agent_manager.rebuild_agents_messages(
                        self.message_handler.streamline_messages
                    )

                    self.message_handler._notify("consolidation_completed", result)

                    if self.message_handler.current_conversation_id:
                        try:
                            self.message_handler.persistent_service.append_conversation_messages(
                                self.message_handler.current_conversation_id,
                                self.message_handler.streamline_messages,
                                True,  # Replace all messages with the consolidated version
                            )
                        except Exception as e:
                            self.message_handler._notify(
                                "error",
                                f"Failed to save consolidated conversation: {str(e)}",
                            )

                    message = (
                        f"Consolidated {result['messages_consolidated']} messages, "
                        f"preserving {result['messages_preserved']} recent messages. "
                        f"Token savings: ~{result['original_token_count'] - result['consolidated_token_count']}"
                    )
                    self.message_handler._notify("system_message", message)
                else:
                    self.message_handler._notify(
                        "system_message",
                        f"Consolidation skipped: {result['reason']}",
                    )

                return CommandResult(handled=True, clear_flag=True)
            else:
                self.message_handler._notify(
                    "error",
                    "Consolidation is only supported with LocalAgent.",
                )
                return CommandResult(handled=False, clear_flag=False)
        except ValueError as e:
            self.message_handler._notify(
                "error",
                f"Invalid consolidation parameter: {str(e)}. Use /consolidate [number]",
            )
            return CommandResult(handled=True, clear_flag=True)
        except Exception as e:
            self.message_handler._notify(
                "error", f"Error during consolidation: {str(e)}"
            )
            return CommandResult(handled=True, clear_flag=True)

    async def _handle_unconsolidate_command(self, user_input: str) -> CommandResult:
        """Handle unconsolidate command to remove last consolidated message."""
        try:
            if isinstance(self.message_handler.agent, LocalAgent):
                consolidator = ConversationConsolidator(self.message_handler.agent.llm)

                result = await consolidator.unconsolidate(
                    self.message_handler.streamline_messages
                )

                if result["success"]:
                    self.message_handler.agent_manager.rebuild_agents_messages(
                        self.message_handler.streamline_messages
                    )

                    self.message_handler._notify("unconsolidation_completed", result)

                    if self.message_handler.current_conversation_id:
                        try:
                            self.message_handler.persistent_service.append_conversation_messages(
                                self.message_handler.current_conversation_id,
                                self.message_handler.streamline_messages,
                                True,
                            )
                        except Exception as e:
                            self.message_handler._notify(
                                "error",
                                f"Failed to save unconsolidated conversation: {str(e)}",
                            )

                    message = (
                        f"Unconsolidated last consolidated message containing "
                        f"{result['messages_restored']} original messages."
                    )
                    self.message_handler._notify("system_message", message)
                else:
                    self.message_handler._notify(
                        "system_message",
                        f"Unconsolidation skipped: {result['reason']}",
                    )

                return CommandResult(handled=True, clear_flag=True)
            else:
                self.message_handler._notify(
                    "error",
                    "Unconsolidation is only supported with LocalAgent.",
                )
                return CommandResult(handled=False, clear_flag=False)
        except Exception as e:
            self.message_handler._notify(
                "error", f"Error during unconsolidation: {str(e)}"
            )
            return CommandResult(handled=True, clear_flag=True)

    async def _handle_mcp_command(self, command: str) -> Tuple[bool, bool]:
        """
        Handle the /mcp command: list prompts or fetch a specific prompt content.

        Returns:
            Tuple of (exit_flag, clear_flag)
        """
        parts = command.strip().split()
        mcp_service: Optional[MCPService] = self.message_handler.mcp_manager.mcp_service
        # /mcp with no args: list all prompts
        if len(parts) == 1:
            prompts = []
            if mcp_service:
                for server_id, prompt_list in mcp_service.server_prompts.items():
                    for prompt in prompt_list:
                        prompt_name = prompt.name
                        if prompt_name:
                            prompts.append(f"{server_id}/{prompt_name}")
            msg = (
                "Available MCP prompts:\n" + "\n".join(prompts)
                if prompts
                else "No MCP prompts found."
            )
            self.message_handler._notify("system_message", msg)
            return False, True
        # /mcp <server_id.prompt_name>: fetch and show the prompt
        elif len(parts) == 2:
            full_name = parts[1]
            if "/" not in full_name:
                self.message_handler._notify(
                    "error", "Please use format: /mcp server_id/prompt_name"
                )
                return False, True
            server_id, prompt_name = full_name.split("/", 1)
            try:
                prompt = await mcp_service.get_prompt(server_id, prompt_name)
                prompt_content = prompt.get("content", [])
                if len(prompt_content) > 0:
                    prompt_text = prompt_content[0].content.text
                    self.message_handler._notify(
                        "mcp_prompt",
                        {"name": prompt_name, "content": f"{prompt_text}"},
                    )
                else:
                    self.message_handler._notify(
                        "error", f"Prompt {server_id}.{prompt_name} not found."
                    )
            except Exception as e:
                self.message_handler._notify(
                    "error", f"Error fetching prompt: {str(e)}"
                )
            return False, True
        else:
            self.message_handler._notify("error", "Usage: /mcp [server_id.prompt_name]")
            return False, True

    def _handle_jump_command(self, command: str) -> bool:
        """Handle the /jump command to rewind conversation to a previous turn."""
        try:
            parts = command.split()
            if len(parts) != 2:
                self.message_handler._notify("error", "Usage: /jump <turn_number>")
                return False

            turn_number = int(parts[1])

            if turn_number < 1 or turn_number > len(
                self.message_handler.conversation_turns
            ):
                self.message_handler._notify(
                    "error",
                    f"Invalid turn number. Available turns: 1-{len(self.message_handler.conversation_turns)}",
                )
                return False

            selected_turn = self.message_handler.conversation_turns[turn_number - 1]

            self.message_handler.streamline_messages = (
                self.message_handler.streamline_messages[: selected_turn.message_index]
            )
            if self.message_handler.current_conversation_id:
                self.message_handler.persistent_service.append_conversation_messages(
                    self.message_handler.current_conversation_id,
                    self.message_handler.streamline_messages,
                    True,
                )

            last_assistant_message = next(
                (
                    msg
                    for msg in reversed(self.message_handler.streamline_messages)
                    if msg.get("role") == "assistant"
                ),
                None,
            )
            if last_assistant_message and last_assistant_message.get("agent", ""):
                self._handle_agent_command(f"/agent {last_assistant_message['agent']}")

            self.message_handler.agent_manager.rebuild_agents_messages(
                self.message_handler.streamline_messages
            )
            self.message_handler.conversation_turns = (
                self.message_handler.conversation_turns[: turn_number - 1]
            )
            self.message_handler.last_assisstant_response_idx = len(
                self.message_handler.agent.history
            )

            self.message_handler._notify(
                "jump_performed",
                {"turn_number": turn_number, "preview": selected_turn.get_preview(100)},
            )

            return True

        except ValueError:
            self.message_handler._notify(
                "error", "Invalid turn number. Please provide a number."
            )
            return False

    def _handle_model_command(self, command: str) -> Tuple[bool, bool]:
        """
        Handle the /model command to switch models or list available models.

        Returns:
            Tuple of (exit_flag, clear_flag)
        """
        model_id = command[7:].strip()
        registry = ModelRegistry.get_instance()
        manager = ServiceManager.get_instance()

        # If no model ID is provided, list available models
        if not model_id:
            models_by_provider = {}
            for provider in registry.get_providers():
                models = registry.get_models_by_provider(provider)
                if models:
                    models_by_provider[provider] = []
                    for model in models:
                        current = (
                            registry.current_model
                            and registry.current_model.id == model.id
                        )
                        models_by_provider[provider].append(
                            {
                                "id": f"{model.provider}/{model.id}",
                                "name": model.name,
                                "description": model.description,
                                "capabilities": model.capabilities,
                                "current": current,
                            }
                        )

            self.message_handler._notify("models_listed", models_by_provider)
            return False, True

        if registry.set_current_model(model_id):
            model = registry.get_current_model()
            if model:
                manager.set_model(model.provider, model.id)

                new_llm_service = manager.get_service(model.provider)

                self.message_handler.agent_manager.update_llm_service(new_llm_service)

                try:
                    config_manager = ConfigManagement()
                    config_manager.set_last_used_model(model_id, model.provider)
                except Exception as e:
                    print(f"Warning: Failed to save last used model: {e}")

                self.message_handler._notify(
                    "model_changed",
                    {"id": model.id, "name": model.name, "provider": model.provider},
                )
            else:
                self.message_handler._notify("error", "Failed to switch model.")
        else:
            self.message_handler._notify("error", f"Unknown model: {model_id}")

        return False, True

    def _handle_agent_command(self, command: str) -> Tuple[bool, str]:
        """
        Handle the /agent command to switch agents or list available agents.

        Returns:
            Tuple of (success, message)
        """
        parts = command.split()

        if len(parts) == 1:
            agents_info = {"current": self.message_handler.agent.name, "available": {}}

            for agent_name, agent in self.message_handler.agent_manager.agents.items():
                agents_info["available"][agent_name] = {
                    "description": agent.description,
                    "current": (
                        self.message_handler.agent
                        and self.message_handler.agent.name == agent_name
                    ),
                }

            self.message_handler._notify("agents_listed", agents_info)
            return True, "Listed available agents"

        agent_name = parts[1]
        old_agent_name = self.message_handler.agent_manager.get_current_agent().name
        if old_agent_name == agent_name:
            return (False, f"Already using {agent_name} agent")
        if self.message_handler.agent_manager.select_agent(agent_name):
            self.message_handler.agent = (
                self.message_handler.agent_manager.get_current_agent()
            )
            old_agent = self.message_handler.agent_manager.get_agent(old_agent_name)
            if old_agent:
                self.message_handler.agent.history = list(old_agent.history)
                old_agent.history = []

            try:
                config_manager = ConfigManagement()
                config_manager.set_last_used_agent(agent_name)
            except Exception as e:
                # Don't fail the command if config save fails, just log it
                print(f"Warning: Failed to save last used agent: {e}")

            self.message_handler._notify("agent_changed", agent_name)
            return True, f"Switched to {agent_name} agent"
        else:
            available_agents = ", ".join(
                self.message_handler.agent_manager.agents.keys()
            )
            self.message_handler._notify(
                "error",
                f"Unknown agent: {agent_name}. Available agents: {available_agents}",
            )
            return (
                False,
                f"Unknown agent: {agent_name}. Available agents: {available_agents}",
            )

    def _handle_file_command(self, user_input: str) -> CommandResult:
        """Handle file command with support for multiple files."""
        # Extract file paths from user input (space-separated)
        file_paths_str: str = user_input[6:].strip()
        file_paths: List[str] = [
            os.path.expanduser(path.strip())
            for path in shlex.split(file_paths_str)
            if path.strip()
        ]

        if not file_paths:
            self.message_handler._notify("error", "No file paths provided")
            return CommandResult(handled=True, clear_flag=True)

        processed_files: List[str] = []
        failed_files: List[str] = []
        all_file_contents: List[Dict[str, str]] = []

        # Process each file
        for file_path in file_paths:
            # self.message_handler._notify("file_processing", {"file_path": file_path})

            # Process file with the file handling service
            if self.message_handler.file_handler is None:
                self.message_handler.file_handler = FileHandler()
            file_content = self.message_handler.file_handler.process_file(file_path)

            # Fallback to llm handle
            if not file_content:
                from AgentCrew.modules.agents.base import MessageType

                file_content = self.message_handler.agent.format_message(
                    MessageType.FileContent, {"file_uri": file_path}
                )

            if file_content:
                all_file_contents.append(file_content)
                processed_files.append(file_path)
                self.message_handler._notify(
                    "file_processed",
                    {
                        "file_path": file_path,
                        "message": file_content,
                    },
                )
            else:
                failed_files.append(file_path)
                self.message_handler._notify(
                    "error",
                    f"Failed to process file {file_path} Or Model is not supported",
                )

        # Add all successfully processed file contents to messages
        if all_file_contents:
            self.message_handler._messages_append(
                {"role": "user", "content": all_file_contents}
            )

            # Notify about overall processing results
            if failed_files:
                self.message_handler._notify(
                    "error", f"Failed to process: {', '.join(failed_files)}"
                )
            self.message_handler._notify(
                "system_message",
                f"✅ Successfully processed {len(processed_files)} files: {', '.join(processed_files)}",
            )

        return CommandResult(handled=True, clear_flag=True)

    def _handle_drop_command(self, user_input: str) -> CommandResult:
        """Handle drop command to remove queued files."""
        # Extract file identifier from user input
        file_path = user_input[6:].strip()  # Remove "/drop " prefix

        if not file_path:
            # Show available files if no argument provided
            if not self.message_handler._queued_attached_files:
                self.message_handler._notify(
                    "error", "No files are currently queued for processing"
                )
                return CommandResult(handled=True, clear_flag=True)

            self.message_handler._notify(
                "system_message",
                "📋 Queued files:\n"
                + "\n".join(self.message_handler._queued_attached_files)
                + "\nUsage: /drop <file_id>",
            )
            return CommandResult(handled=True, clear_flag=True)

        try:
            # Get the file command and parse its file paths
            try:
                self.message_handler._queued_attached_files.remove(file_path)
            except Exception:
                self.message_handler._notify(
                    "error", f"Cannot unqueue file: {file_path}"
                )

            # Update or remove the command
            self.message_handler._notify(
                "system_message", f"🗑️ Removed file from queue: {file_path}"
            )

            # Notify about the file being removed for UI updates
            self.message_handler._notify("file_dropped", {"file_path": file_path})

            return CommandResult(handled=True, clear_flag=True)

        except ValueError as e:
            self.message_handler._notify("error", f"Invalid file ID format: {str(e)}")
            return CommandResult(handled=True, clear_flag=True)
        except Exception as e:
            self.message_handler._notify("error", f"Error removing file: {str(e)}")
            return CommandResult(handled=True, clear_flag=True)

    async def _handle_voice_command(self, command: str) -> CommandResult:
        """Handle /voice command to start voice recording."""
        try:
            # Check if already recording
            if self.message_handler.voice_service is None:
                self.message_handler._notify(
                    "error",
                    "Voice service not available. Set ELEVENLABS_API_KEY environment variable.",
                )
                return CommandResult(handled=True, clear_flag=True)

            if self.message_handler.voice_service.is_recording():
                self.message_handler._notify(
                    "error",
                    "Already recording. Use /end_voice to stop current recording.",
                )
                return CommandResult(handled=True, clear_flag=True)

            # Start recording
            result = self.message_handler.voice_service.start_voice_recording()

            if result["success"]:
                self.message_handler._notify("voice_recording_started", None)
                self.message_handler._notify(
                    "system_message",
                    "🎤 Recording started. Press Enter to stop.",
                )
            else:
                self.message_handler._notify("error", result["error"])

            return CommandResult(handled=True, clear_flag=True)

        except Exception as e:
            self.message_handler._notify("error", f"Voice command failed: {str(e)}")
            return CommandResult(handled=True, clear_flag=True)

    async def _handle_end_voice_command(self, command: str) -> CommandResult:
        """Handle /end_voice command to stop recording and transcribe."""
        try:
            # Check if voice service exists and is recording
            if self.message_handler.voice_service is None:
                self.message_handler._notify(
                    "error",
                    "No voice service initialized. Use /voice to start recording.",
                )
                return CommandResult(handled=True, clear_flag=True)

            if not self.message_handler.voice_service.is_recording():
                self.message_handler._notify("error", "No recording in progress.")
                return CommandResult(handled=True, clear_flag=True)

            # Stop recording
            self.message_handler._notify("voice_recording_stopping", None)
            stop_result = self.message_handler.voice_service.stop_voice_recording()

            if not stop_result["success"]:
                self.message_handler._notify("error", stop_result["error"])
                return CommandResult(handled=True, clear_flag=True)

            # Transcribe
            self.message_handler._notify("system_message", "🔄 Transcribing audio...")

            transcribe_result = await self.message_handler.voice_service.speech_to_text(
                stop_result["audio_data"], stop_result["sample_rate"]
            )
            transcribed_text = None

            if transcribe_result["success"]:
                transcribed_text = transcribe_result["text"]
                confidence = transcribe_result.get("confidence", 1.0)

                # Notify about transcription
                self.message_handler._notify(
                    "system_message",
                    f"✅ Transcribed (confidence: {confidence:.0%}): {transcribed_text}",
                )

                # Process as regular user input if there's text
                if transcribed_text.strip():
                    await self.message_handler.process_user_input(transcribed_text)
                else:
                    self.message_handler._notify(
                        "system_message", "No speech detected in the recording."
                    )

            else:
                self.message_handler._notify("error", transcribe_result["error"])

            self.message_handler._notify("voice_recording_completed", transcribed_text)
            return CommandResult(handled=True, clear_flag=False)

        except Exception as e:
            self.message_handler._notify("error", f"End voice command failed: {str(e)}")
            self.message_handler._notify("voice_recording_completed", None)
            return CommandResult(handled=True, clear_flag=True)



from typing import Optional, Tuple, Any
import threading
import queue
import numpy as np
import sounddevice as sd
import pyaudio
from AgentCrew.modules import logger


class AudioHandler:
    """Handles audio recording and playback operations."""

    def __init__(self):
        """Initialize audio handler."""
        self.recording = False
        self.recording_thread = None
        self.audio_queue = queue.Queue()
        self.pyaudio = pyaudio.PyAudio()
        self.current_sample_rate = 44100

    def start_recording(self, sample_rate: int = 44100) -> None:
        """
        Start recording audio in a separate thread.

        Args:
            sample_rate: Sample rate for recording
        """
        if self.recording:
            logger.warning("Recording already in progress")
            return

        self.recording = True
        self.current_sample_rate = sample_rate
        self.audio_queue.queue.clear()  # Clear any previous data

        self.recording_thread = threading.Thread(
            target=self._recording_worker, args=(sample_rate,), daemon=True
        )
        self.recording_thread.start()
        logger.info("Recording started")

    def stop_recording(self) -> Tuple[Optional[Any], int]:
        """
        Stop recording and return the recorded audio.

        Returns:
            Tuple of (audio_data, sample_rate) or (None, 0) if no data
        """
        if not self.recording:
            logger.warning("No recording in progress")
            return None, 0

        self.recording = False

        # Wait for recording thread to finish
        if self.recording_thread:
            self.recording_thread.join(timeout=1.0)

        # Collect all recorded frames
        frames = []
        while not self.audio_queue.empty():
            try:
                frames.append(self.audio_queue.get_nowait())
            except queue.Empty:
                break

        if frames:
            audio_data = np.concatenate(frames, axis=0).flatten()
            logger.info(
                f"Recording stopped. Captured {len(audio_data) / self.current_sample_rate:.2f} seconds"
            )
            return audio_data, self.current_sample_rate
        else:
            logger.warning("No audio data captured")
            return None, 0

    def _recording_worker(self, sample_rate: int):
        """Worker thread for continuous recording."""
        try:

            def callback(indata, frames, time, status):
                if status:
                    logger.warning(f"Recording status: {status}")
                if self.recording:
                    self.audio_queue.put(indata.copy())

            with sd.InputStream(
                samplerate=sample_rate,
                channels=1,
                callback=callback,
                dtype="float32",
                blocksize=1024,
            ):
                while self.recording:
                    sd.sleep(100)  # Sleep for 100ms chunks

        except Exception as e:
            logger.error(f"Recording error: {str(e)}")
            self.recording = False

    def is_recording(self) -> bool:
        """Check if currently recording."""
        return self.recording

    def __del__(self):
        """Cleanup PyAudio."""
        try:
            if self.recording:
                self.stop_recording()
            self.pyaudio.terminate()
        except Exception:
            pass



from .token_usage import TokenUsageWidget
from .system_message import SystemMessageWidget
from .message_bubble import MessageBubble
from .history_sidebar import ConversationSidebar, ConversationLoader
from .tool_widget import ToolWidget

__all__ = [
    "TokenUsageWidget",
    "SystemMessageWidget",
    "MessageBubble",
    "ConversationSidebar",
    "ConversationLoader",
    "ToolWidget",
]



from AgentCrew.modules import logger
from typing import Dict, Any, List, Optional, Callable
from mcp import ClientSession, StdioServerParameters
from mcp.types import Prompt, ContentBlock, TextContent, ImageContent, Tool
from mcp.client.stdio import stdio_client
from mcp.client.streamable_http import streamablehttp_client
from AgentCrew.modules.agents import AgentManager, LocalAgent
from AgentCrew.modules.tools.registry import ToolRegistry
from .config import MCPServerConfig
import asyncio
import threading
from AgentCrew.modules import FileLogIO

# Initialize the logger
mcp_log_io = FileLogIO("mcpclient_agentcrew")


class MCPService:
    """Service for interacting with Model Context Protocol (MCP) servers."""

    def __init__(self):
        """Initialize the MCP service."""
        self.sessions: Dict[str, ClientSession] = {}
        self.connected_servers: Dict[str, bool] = {}
        self.tools_cache: Dict[str, Dict[str, Any]] = {}
        self.loop = asyncio.new_event_loop()
        self._server_connection_tasks: Dict[str, asyncio.Task] = {}
        self._server_shutdown_events: Dict[str, asyncio.Event] = {}
        self.server_prompts: Dict[str, List[Prompt]] = {}

    async def _manage_single_connection(
        self, server_config: MCPServerConfig, agent_name: Optional[str] = None
    ):
        """Manages the lifecycle of a single MCP server connection."""
        server_name = server_config.name
        combined_server_id = self._get_server_id_format(server_name, agent_name)
        shutdown_event = asyncio.Event()
        self._server_shutdown_events[combined_server_id] = shutdown_event
        logger.info(f"MCPService: Starting connection management for {server_name}")

        try:
            if server_config.streaming_server:
                # Import here to avoid import errors if not available

                logger.info(
                    f"MCPService: Using streaming HTTP client for {server_name}"
                )

                # Prepare headers for the streamable HTTP client
                headers = server_config.headers if server_config.headers else {}

                async with streamablehttp_client(
                    server_config.url, headers=headers
                ) as (
                    read_stream,
                    write_stream,
                    _,
                ):
                    logger.info(
                        f"MCPService: streamablehttp_client established for {server_name}"
                    )
                    async with ClientSession(read_stream, write_stream) as session:
                        logger.info(
                            f"MCPService: ClientSession established for {server_name}"
                        )
                        server_info = await session.initialize()
                        self.sessions[combined_server_id] = session
                        self.connected_servers[combined_server_id] = True
                        logger.info(
                            f"MCPService: {server_name} connected. Registering tools..."
                        )

                        if agent_name:
                            await self.register_server_tools(
                                combined_server_id, server_name, agent_name
                            )
                        else:
                            for agent_name in server_config.enabledForAgents:
                                await self.register_server_tools(
                                    combined_server_id, server_name, agent_name
                                )

                        if server_info.capabilities.prompts:
                            prompts = await self.sessions[
                                combined_server_id
                            ].list_prompts()
                            self.server_prompts[server_name] = prompts.prompts

                        logger.info(
                            f"MCPService: {server_name} setup complete. Waiting for shutdown signal."
                        )
                        await shutdown_event.wait()
            else:
                # Original stdio client logic
                server_params = StdioServerParameters(
                    command=server_config.command,
                    args=server_config.args,
                    env=server_config.env,
                )

                async with stdio_client(server_params, errlog=mcp_log_io) as (
                    read_stream,
                    write_stream,
                ):
                    logger.info(
                        f"MCPService: stdio_client established for {server_name}"
                    )
                    async with ClientSession(read_stream, write_stream) as session:
                        logger.info(
                            f"MCPService: ClientSession established for {server_name}"
                        )
                        server_info = await session.initialize()
                        self.sessions[combined_server_id] = session
                        self.connected_servers[combined_server_id] = (
                            True  # Mark as connected before tool registration
                        )
                        logger.info(
                            f"MCPService: {combined_server_id} connected. Registering tools..."
                        )

                        if agent_name:
                            await self.register_server_tools(
                                combined_server_id, server_name, agent_name
                            )
                        else:
                            for agent_name in server_config.enabledForAgents:
                                await self.register_server_tools(
                                    combined_server_id, server_name, agent_name
                                )

                        if server_info.capabilities.prompts:
                            prompts = await self.sessions[
                                combined_server_id
                            ].list_prompts()
                            self.server_prompts[server_name] = prompts.prompts

                        logger.info(
                            f"MCPService: {server_name} setup complete. Waiting for shutdown signal."
                        )
                        await shutdown_event.wait()

        except asyncio.CancelledError:
            logger.info(f"MCPService: Connection task for {server_name} was cancelled.")
        except Exception:
            logger.exception(
                f"MCPService: Error in connection management for '{server_name}'"
            )
            if agent_name:
                agent_manager = AgentManager.get_instance()
                agent = agent_manager.get_local_agent(agent_name)
                if agent:
                    agent.is_tool_ready = True
        finally:
            logger.info(f"MCPService: Cleaning up connection for {server_name}.")
            self.sessions.pop(server_name, None)
            self.connected_servers.pop(server_name, False)
            self.tools_cache.pop(server_name, None)
            self._server_shutdown_events.pop(server_name, None)
            logger.info(f"MCPService: Cleanup for {server_name} complete.")

    def _get_server_id_format(
        self, server_name: str, agent_name: Optional[str] = None
    ) -> str:
        """Format server ID with optional agent name prefix."""
        return f"{agent_name}__{server_name}" if agent_name else server_name

    async def start_server_connection_management(
        self, server_config: MCPServerConfig, agent_name: Optional[str] = None
    ):
        """Starts and manages the connection for a single MCP server."""
        combined_server_id = self._get_server_id_format(server_config.name, agent_name)
        if (
            combined_server_id in self._server_connection_tasks
            and not self._server_connection_tasks[combined_server_id].done()
        ):
            logger.info(
                f"MCPService: Connection management for {combined_server_id} already in progress."
            )
            return

        if agent_name:
            agent_manager = AgentManager.get_instance()
            agent = agent_manager.get_local_agent(agent_name)
            if agent:
                agent.is_tool_ready = False
        logger.info(
            f"MCPService: Creating task for _manage_single_connection for {combined_server_id}"
        )
        if self.loop.is_closed():
            logger.warning(
                "MCPService: Loop is closed, cannot create task for server connection."
            )
            return
        task = self.loop.create_task(
            self._manage_single_connection(server_config, agent_name)
        )
        self._server_connection_tasks[combined_server_id] = task

    async def shutdown_all_server_connections(self, agent_name: Optional[str] = None):
        """Signals all active server connections to shut down and waits for them."""
        logger.info("MCPService: Shutting down all server connections...")
        active_tasks = []
        for server_id, event in list(self._server_shutdown_events.items()):
            if agent_name and agent_name not in server_id:
                continue  # Skip servers not matching the agent name
            extracted_server_id = server_id.replace(
                f"{agent_name}__", ""
            )  # Remove prefix if present
            await self.deregister_server_tools(extracted_server_id, agent_name)
            logger.info(f"MCPService: Signaling shutdown for {server_id}")
            event.set()
            if server_id in self._server_connection_tasks:
                task = self._server_connection_tasks[server_id]
                if task and not task.done():
                    active_tasks.append(task)

        if active_tasks:
            logger.info(
                f"MCPService: Waiting for {len(active_tasks)} connection tasks to complete..."
            )
            await asyncio.gather(*active_tasks, return_exceptions=True)

        self._server_connection_tasks.clear()
        logger.info("MCPService: All server connections shut down process initiated.")

    async def shutdown_single_server_connection(self, server_id: str):
        """Signals a specific server connection to shut down and waits for it."""
        logger.info(f"MCPService: Shutting down connection for server {server_id}...")
        if server_id in self._server_shutdown_events:
            event = self._server_shutdown_events[server_id]
            event.set()
            logger.info(f"MCPService: Shutdown signal sent to {server_id}.")

            task = self._server_connection_tasks.get(server_id)
            if task and not task.done():
                logger.info(
                    f"MCPService: Waiting for connection task of {server_id} to complete..."
                )
                try:
                    await asyncio.wait_for(
                        task, timeout=10.0
                    )  # Wait for task to finish
                    logger.info(
                        f"MCPService: Connection task for {server_id} completed."
                    )
                except asyncio.TimeoutError:
                    logger.warning(
                        f"MCPService: Timeout waiting for {server_id} connection task to complete. It might be stuck."
                    )
                    task.cancel()  # Force cancel if it didn't finish
                except Exception as e:
                    logger.error(f"MCPService: Error waiting for {server_id} task: {e}")
            else:
                logger.info(
                    f"MCPService: No active task found for {server_id} or task already done."
                )
        else:
            logger.warning(
                f"MCPService: No shutdown event found for server {server_id}. It might not be running or already shut down."
            )

        # Clean up entries related to this server, though _manage_single_connection's finally should handle most
        self._server_connection_tasks.pop(server_id, None)
        logger.info(
            f"MCPService: Shutdown process for {server_id} initiated/completed."
        )

    async def register_server_tools(
        self, combined_server_id: str, server_name: str, agent_name: str
    ) -> None:
        """
        Register all tools from a connected server.

        Args:
            server_id: ID of the server to register tools from
        """

        if combined_server_id not in self.sessions or not self.connected_servers.get(
            combined_server_id
        ):
            logger.warning(
                f"Cannot register tools: Server '{combined_server_id}' is not connected"
            )
            return

        try:
            response = await self.sessions[combined_server_id].list_tools()

            # Cache tools
            self.tools_cache[combined_server_id] = {
                tool.name: tool for tool in response.tools
            }

            if agent_name:
                agent_manager = AgentManager.get_instance()
                registry = agent_manager.get_local_agent(agent_name)
            else:
                # Register each tool with the tool registry
                registry = ToolRegistry.get_instance()
            for tool in response.tools:
                # Create namespaced tool definition
                def tool_definition_factory(tool_info=tool, srv_id=server_name):
                    def get_definition(provider=None):
                        return self._format_tool_definition(tool_info, srv_id, provider)

                    return get_definition

                # Create tool handler
                handler_factory = self._create_tool_handler(
                    combined_server_id, tool.name
                )

                # Register the tool
                if registry:
                    registry.register_tool(
                        tool_definition_factory(), handler_factory, self
                    )
            if isinstance(registry, LocalAgent):
                registry.is_tool_ready = True

        except Exception:
            logger.exception(
                f"Error registering tools from server '{combined_server_id}'"
            )
            self.connected_servers[combined_server_id] = False

    async def deregister_server_tools(
        self, server_name: str, agent_name: Optional[str] = None
    ):
        agent_manager = AgentManager.get_instance()
        if agent_name:
            local_agent = agent_manager.get_local_agent(agent_name)
            if not local_agent:
                return
            combined_server_id = self._get_server_id_format(server_name, agent_name)
            if server_name in self.tools_cache:
                for tool_name in self.tools_cache[combined_server_id].keys():
                    if local_agent.is_active:
                        local_agent.deactivate()
                    if (
                        f"{server_name}_{tool_name}"
                        in local_agent.tool_definitions.keys()
                    ):
                        del local_agent.tool_definitions[f"{server_name}_{tool_name}"]
        else:
            for agent_name in agent_manager.agents.keys():
                local_agent = agent_manager.get_local_agent(agent_name)
                if not local_agent:
                    continue

                combined_server_id = self._get_server_id_format(server_name, agent_name)
                if combined_server_id in self.tools_cache:
                    for tool_name in self.tools_cache[combined_server_id].keys():
                        if local_agent.is_active:
                            local_agent.deactivate()
                        if (
                            f"{server_name}_{tool_name}"
                            in local_agent.tool_definitions.keys()
                        ):
                            del local_agent.tool_definitions[
                                f"{server_name}_{tool_name}"
                            ]

    def _format_tool_definition(
        self, tool: Tool, server_id: str, provider: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Format a tool definition for the tool registry.

        Args:
            tool: Tool information from the server
            server_id: ID of the server the tool belongs to
            provider: LLM provider to format for (if None, uses default format)

        Returns:
            Formatted tool definition
        """
        # Create namespaced tool name
        namespaced_name = f"{server_id}_{tool.name}"

        # Format for different providers
        if provider == "claude":
            return {
                "name": namespaced_name,
                "description": tool.description,
                "input_schema": tool.inputSchema,
            }
        else:  # Default format (OpenAI-compatible)
            return {
                "type": "function",
                "function": {
                    "name": namespaced_name,
                    "description": tool.description,
                    "parameters": tool.inputSchema,
                },
            }

    def _run_async(self, coro):
        """Helper method to run coroutines in the service's event loop"""
        if self.loop.is_closed():
            self.loop = asyncio.new_event_loop()
        return asyncio.run_coroutine_threadsafe(coro, self.loop).result()

    def _create_tool_handler(self, server_id: str, tool_name: str) -> Callable:
        """
        Create an asynchronous handler function for a tool.

        Args:
            server_id: ID of the server the tool belongs to
            tool_name: Name of the tool

        Returns:
            Asynchronous handler function for the tool
        """

        # This is the factory that the ToolRegistry expects.
        # It should return the actual callable that will be invoked by the agent.
        def handler_factory(
            service_instance=None,
        ):  # service_instance will be self (MCPService)
            # This is the actual async handler the agent will await.
            def actual_tool_executor(
                **params,
            ) -> list[Dict[str, Any]]:
                if server_id not in self.sessions or not self.connected_servers.get(
                    server_id
                ):
                    raise Exception(
                        f"Cannot call tool: Server '{server_id}' is not connected"
                    )

                # The agent framework should be awaiting this coroutine.
                # The call to session.call_tool is already async.
                session = self.sessions[server_id]
                result = self._run_async(session.call_tool(tool_name, params))
                return self._format_contents(result.content)

            return actual_tool_executor  # Return the async function

        return handler_factory  # Return the factory

    async def list_tools(self, server_id: Optional[str] = None) -> List[Dict[str, Any]]:
        """
        List all available tools from a connected MCP server or all servers.

        Args:
            server_id: Optional ID of the server to list tools from. If None, lists tools from all servers.

        Returns:
            List of tools with their metadata
        """
        if server_id is not None:
            if server_id not in self.sessions or not self.connected_servers.get(
                server_id
            ):
                return []

            try:
                response = await self.sessions[server_id].list_tools()
                self.tools_cache[server_id] = {
                    tool.name: tool for tool in response.tools
                }

                return [
                    {
                        "name": f"{server_id}.{tool.name}",
                        "description": tool.description,
                        "input_schema": tool.inputSchema,
                    }
                    for tool in response.tools
                ]
            except Exception:
                logger.exception(f"Error listing tools from server '{server_id}'")
                self.connected_servers[server_id] = False
                return []
        else:
            # List tools from all connected servers
            all_tools = []
            for srv_id in list(self.sessions.keys()):
                all_tools.extend(await self.list_tools(srv_id))
            return all_tools

    async def get_prompt(
        self,
        server_id: str,
        prompt_name: str,
        arguments: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Get a specific prompt from a connected MCP server.

        Args:
            server_id: ID of the server to get the prompt from
            prompt_name: Name of the prompt to retrieve

        Returns:
            Prompt object if found, None otherwise
        """

        if server_id not in self.sessions or not self.connected_servers.get(server_id):
            return {
                "content": f"Cannot call tool: Server '{server_id}' is not connected",
                "status": "error",
            }

        try:
            session = self.sessions[server_id]
            result = self._run_async(session.get_prompt(prompt_name, arguments))
            return {"content": result.messages, "status": "success"}
        except Exception as e:
            logger.error(
                f"Error retrieving prompt '{prompt_name}' from server '{server_id}': {str(e)}"
            )
            self.connected_servers[server_id] = False
            return {
                "content": f"Error calling tool '{prompt_name}' on server '{server_id}': {str(e)}",
                "status": "error",
            }

    def _format_contents(self, content: List[ContentBlock]) -> List[Dict[str, Any]]:
        result = []
        for c in content:
            if isinstance(c, TextContent):
                result.append(
                    {
                        "type": "text",
                        "text": c.text,
                    }
                )
            elif isinstance(c, ImageContent):
                result.append(
                    {
                        "type": "image_url",
                        "image_url": {"url": f"data:{c.mimeType};base64,{c.data}"},
                    }
                )

        return result

    async def call_tool(
        self, server_id: str, tool_name: str, tool_args: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Call a tool on an MCP server.

        Args:
            server_id: ID of the server to call the tool on
            tool_name: Name of the tool to call
            tool_args: Arguments to pass to the tool

        Returns:
            Dict containing the tool's response
        """
        if server_id not in self.sessions or not self.connected_servers.get(server_id):
            return {
                "content": f"Cannot call tool: Server '{server_id}' is not connected",
                "status": "error",
            }

        if server_id not in self.tools_cache or tool_name not in self.tools_cache.get(
            server_id, {}
        ):
            # Refresh tools cache
            await self.list_tools(server_id)
            if (
                server_id not in self.tools_cache
                or tool_name not in self.tools_cache.get(server_id, {})
            ):
                return {
                    "content": f"Tool '{tool_name}' is not available on server '{server_id}'",
                    "status": "error",
                }

        try:
            result = await self.sessions[server_id].call_tool(tool_name, tool_args)
            return {
                "content": self._format_contents(result.content),
                "status": "success",
            }
        except Exception as e:
            self.connected_servers[server_id] = False
            return {
                "content": f"Error calling tool '{tool_name}' on server '{server_id}': {str(e)}",
                "status": "error",
            }

    def start(self):
        """Start the service's event loop in a separate thread"""

        def run_loop():
            logger.info("MCPService: Event loop thread started.")
            asyncio.set_event_loop(self.loop)
            try:
                self.loop.run_forever()
            finally:
                logger.info("MCPService: Event loop stopping...")
                # This block executes when loop.stop() is called or run_forever() exits.
                # Attempt to cancel any remaining tasks.
                # shutdown_all_server_connections should ideally handle most task terminations gracefully.
                try:
                    all_tasks = asyncio.all_tasks(loop=self.loop)
                    # Exclude the current task if this finally block is run by a task on the loop
                    current_task = (
                        asyncio.current_task(loop=self.loop)
                        if self.loop.is_running()
                        else None
                    )  # Check if loop is running

                    tasks_to_cancel = [
                        t for t in all_tasks if t is not current_task and not t.done()
                    ]
                    if tasks_to_cancel:
                        logger.info(
                            f"MCPService: Cancelling {len(tasks_to_cancel)} outstanding tasks in event loop thread."
                        )
                        for task in tasks_to_cancel:
                            task.cancel()
                        # Give tasks a chance to process cancellation
                        # This needs to run on the loop, but run_forever has exited.
                        # We can run_until_complete for these specific tasks.
                        if self.loop.is_running():  # Should not be, but as a safeguard
                            self.loop.run_until_complete(
                                asyncio.gather(*tasks_to_cancel, return_exceptions=True)
                            )
                        else:  # If loop not running, create a temporary runner for cleanup

                            async def gather_cancel_tasks():
                                await asyncio.gather(
                                    *tasks_to_cancel, return_exceptions=True
                                )

                            self.loop.run_until_complete(gather_cancel_tasks())

                except RuntimeError as e:
                    logger.error(
                        f"MCPService: Runtime error during task cancellation in run_loop finally: {e}"
                    )
                except Exception as e_final:
                    logger.error(
                        f"MCPService: General error during task cancellation in run_loop finally: {e_final}"
                    )

                if not self.loop.is_closed():
                    self.loop.close()
                logger.info("MCPService: Event loop thread stopped and loop closed.")

        self.loop_thread = threading.Thread(target=run_loop, daemon=True)
        self.loop_thread.start()

    def stop(self):
        """Stop the service's event loop"""
        logger.info("MCPService: Stopping event loop...")
        if hasattr(self, "loop") and self.loop and not self.loop.is_closed():
            if self.loop.is_running():
                logger.info("MCPService: Requesting event loop to stop.")
                self.loop.call_soon_threadsafe(self.loop.stop)
            # The finally block in run_loop should handle task cleanup and closing the loop.
        else:
            logger.info(
                "MCPService: Loop not available or already closed when stop() called."
            )

        if hasattr(self, "loop_thread") and self.loop_thread.is_alive():
            logger.info("MCPService: Waiting for event loop thread to join...")
            self.loop_thread.join(timeout=10)
            if self.loop_thread.is_alive():
                logger.warning(
                    "MCPService: Event loop thread did not join in time. Loop might be stuck or tasks not yielding."
                )
        else:
            logger.info(
                "MCPService: Loop thread not available or not alive when stop() called."
            )

        # Fallback: Ensure loop is closed if thread exited but loop wasn't closed by run_loop's finally
        if hasattr(self, "loop") and self.loop and not self.loop.is_closed():
            logger.warning(
                "MCPService: Loop was not closed by thread's run_loop, attempting to close now."
            )
            # This is a fallback, ideally run_loop's finally handles this.
            try:
                # Minimal cleanup if loop is in a weird state
                # Ensure all tasks are finished or cancelled before closing loop
                if (
                    not self.loop.is_running()
                ):  # If not running, we can try to run_until_complete for cleanup
                    tasks = [
                        t for t in asyncio.all_tasks(loop=self.loop) if not t.done()
                    ]
                    if tasks:
                        logger.info(
                            f"MCPService: Running {len(tasks)} pending tasks to completion before closing loop in stop()."
                        )

                        async def finalize_tasks():
                            await asyncio.gather(*tasks, return_exceptions=True)

                        self.loop.run_until_complete(finalize_tasks())
            except (
                RuntimeError
            ) as e:  # e.g. "cannot call run_until_complete() on a running loop"
                logger.error(
                    f"MCPService: Runtime error during final loop cleanup in stop(): {e}"
                )
            except Exception as e_final_stop:
                logger.error(
                    f"MCPService: General error during final loop cleanup in stop(): {e_final_stop}"
                )
            finally:
                if not self.loop.is_closed():  # Check again before closing
                    self.loop.close()
                    logger.info("MCPService: Loop closed in stop() fallback.")

        logger.info("MCPService: Stop process complete.")



from .local_agent import LocalAgent
from .base import BaseAgent
from .manager import AgentManager
from .remote_agent import RemoteAgent

__all__ = ["AgentManager", "LocalAgent", "BaseAgent", "RemoteAgent"]



import json
import os
from dataclasses import dataclass
from typing import Dict, List, Optional
from AgentCrew.modules import logger


@dataclass
class MCPServerConfig:
    """Configuration for an MCP server."""

    name: str
    command: str
    args: List[str]
    enabledForAgents: List[str]
    env: Optional[Dict[str, str]] = None
    streaming_server: bool = False
    url: str = ""
    headers: Optional[Dict[str, str]] = None


class MCPConfigManager:
    """Manager for MCP server configurations."""

    def __init__(self, config_path: Optional[str] = None):
        """
        Initialize the configuration manager.

        Args:
            config_path: Path to the configuration file. If None, uses the default path.
        """
        self.config_path = config_path or os.environ.get(
            "MCP_CONFIG_PATH",
            os.path.join(
                os.path.dirname(
                    os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
                ),
                "mcp_servers.json",
            ),
        )
        self.configs: Dict[str, MCPServerConfig] = {}

    def load_config(self) -> Dict[str, MCPServerConfig]:
        """
        Load server configurations from the config file.

        Returns:
            Dictionary of server configurations keyed by server ID.
        """
        try:
            if not os.path.exists(self.config_path):
                # Create default config directory if it doesn't exist
                os.makedirs(os.path.dirname(self.config_path), exist_ok=True)
                # Create empty config file
                with open(self.config_path, "w") as f:
                    json.dump({}, f)
                return {}

            with open(self.config_path, "r") as f:
                config_data = json.load(f)

            self.configs = {}
            for server_id, config in config_data.items():
                self.configs[server_id] = MCPServerConfig(
                    name=config.get("name", server_id),
                    command=config.get("command", ""),
                    args=config.get("args", []),
                    env=config.get("env"),
                    enabledForAgents=config.get("enabledForAgents", []),
                    streaming_server=config.get("streaming_server", False),
                    url=config.get("url", ""),
                    headers=config.get("headers"),
                )

            return self.configs
        except Exception as e:
            logger.error(f"Error loading MCP configuration: {e}")
            return {}

    def get_enabled_servers(
        self, agent_name: Optional[str] = None
    ) -> Dict[str, MCPServerConfig]:
        """
        Get all enabled server configurations.

        Returns:
            Dictionary of enabled server configurations.
        """
        if agent_name:
            return {
                server_id: config
                for server_id, config in self.configs.items()
                if agent_name in config.enabledForAgents
            }

        return {
            server_id: config
            for server_id, config in self.configs.items()
            if len(config.enabledForAgents) > 0
        }



from .chroma_service import ChromaMemoryService
from .base_service import BaseMemoryService
from .tool import (
    get_memory_retrieve_tool_definition,
    get_memory_retrieve_tool_handler,
    get_memory_forget_tool_definition,
    get_memory_forget_tool_handler,
)
from .context_persistent import ContextPersistenceService


try:
    from .mem0_service import Mem0MemoryService

    __all__ = [
        "ChromaMemoryService",
        "BaseMemoryService",
        "Mem0MemoryService",
        "get_memory_retrieve_tool_definition",
        "get_memory_retrieve_tool_handler",
        "get_memory_forget_tool_definition",
        "get_memory_forget_tool_handler",
        "ContextPersistenceService",
    ]

except ModuleNotFoundError:
    __all__ = [
        "ChromaMemoryService",
        "BaseMemoryService",
        "get_memory_retrieve_tool_definition",
        "get_memory_retrieve_tool_handler",
        "get_memory_forget_tool_definition",
        "get_memory_forget_tool_handler",
        "ContextPersistenceService",
    ]



import nest_asyncio
from AgentCrew.modules.config import ConfigManagement
import click
import importlib
import os
import sys
import traceback
import json
import requests
import time
import subprocess
import platform
from AgentCrew.modules.console import ConsoleUI
from AgentCrew.modules.gui import ChatWindow
from AgentCrew.modules.chat import MessageHandler
from AgentCrew.modules.web_search import TavilySearchService
from AgentCrew.modules.clipboard import ClipboardService
from AgentCrew.modules.memory import (
    ChromaMemoryService,
    ContextPersistenceService,
)
from AgentCrew.modules.code_analysis import CodeAnalysisService
from AgentCrew.modules.llm.service_manager import ServiceManager
from AgentCrew.modules.llm.model_registry import ModelRegistry
from AgentCrew.modules.agents import AgentManager, LocalAgent, RemoteAgent
from AgentCrew.modules.image_generation import ImageGenerationService
from PySide6.QtWidgets import QApplication

nest_asyncio.apply()


@click.group()
def cli():
    """SwissKnife - AI Assistant and Agent Framework"""
    pass


def cli_prod():
    from AgentCrew.modules import FileLogIO

    sys.stderr = FileLogIO()

    os.environ["AGENTCREW_LOG_PATH"] = os.path.expanduser("~/.AgentCrew/logs")
    os.environ["MEMORYDB_PATH"] = os.path.expanduser("~/.AgentCrew/memorydb")
    os.environ["MCP_CONFIG_PATH"] = os.path.expanduser("~/.AgentCrew/mcp_servers.json")
    os.environ["SW_AGENTS_CONFIG"] = os.path.expanduser("~/.AgentCrew/agents.toml")
    os.environ["PERSISTENCE_DIR"] = os.path.expanduser("~/.AgentCrew/persistents")
    os.environ["AGENTCREW_CONFIG_PATH"] = os.path.expanduser("~/.AgentCrew/config.json")

    check_and_update()
    cli()  # Delegate to main CLI function


def load_api_keys_from_config():
    """Loads API keys from the global config file and sets them as environment variables,
    prioritizing them over any existing environment variables."""

    config_file_path = os.getenv("AGENTCREW_CONFIG_PATH")
    if not config_file_path:
        # Default for when AGENTCREW_CONFIG_PATH is not set (e.g. dev mode, not using cli_prod)
        config_file_path = "./config.json"
    config_file_path = os.path.expanduser(config_file_path)

    api_keys_config = {}
    if os.path.exists(config_file_path):
        try:
            with open(config_file_path, "r", encoding="utf-8") as f:
                loaded_config = json.load(f)
                if isinstance(loaded_config, dict) and isinstance(
                    loaded_config.get("api_keys"), dict
                ):
                    api_keys_config = loaded_config["api_keys"]
                else:
                    click.echo(
                        f"⚠️  API keys in {config_file_path} are not in the expected format.",
                        err=True,
                    )
        except json.JSONDecodeError:
            click.echo(f"⚠️  Error decoding API keys from {config_file_path}.", err=True)
        except Exception as e:
            click.echo(
                f"⚠️  Could not load API keys from {config_file_path}: {e}", err=True
            )

    keys_to_check = [
        "ANTHROPIC_API_KEY",
        "GEMINI_API_KEY",
        "OPENAI_API_KEY",
        "GROQ_API_KEY",
        "DEEPINFRA_API_KEY",
        "GITHUB_COPILOT_API_KEY",
        "TAVILY_API_KEY",
        "VOYAGE_API_KEY",
        "ELEVENLABS_API_KEY",
    ]

    for key_name in keys_to_check:
        if key_name in api_keys_config and api_keys_config[key_name]:
            # Prioritize config file over existing environment variables
            os.environ[key_name] = str(api_keys_config[key_name]).strip()


def check_and_update():
    """Check for updates against the GitHub repository and run update command if needed"""
    try:
        # Get current version from __version__ or a version file
        current_version = get_current_version()

        # Get latest version from GitHub API
        latest_version = get_latest_github_version()

        if not current_version or not latest_version:
            click.echo("⚠️ Could not determine version information", err=True)
            return

        click.echo(f"Current version: {current_version}")
        click.echo(f"Latest version: {latest_version}")

        if version_is_older(current_version, latest_version):
            # Add user confirmation prompt
            if click.confirm(
                "🔄 New version available! Do you want to update now?", default=False
            ):
                click.echo("🔄 Starting update...")
                run_update_command()
                sys.exit(0)  # Exit after update command
            else:
                click.echo("⏭️ Skipping update. Starting application...")
        else:
            click.echo("✅ You are running the latest version")

    except Exception as e:
        click.echo(f"❌ Update check failed: {str(e)}", err=True)


def get_current_version():
    """Get the current version of AgentCrew"""
    try:
        # Try to get version from package __version__ attribute
        import AgentCrew

        if hasattr(AgentCrew, "__version__"):
            return AgentCrew.__version__

        return None
    except Exception:
        return None


def get_latest_github_version():
    """Get the latest version from GitHub repository tags"""
    try:
        # Use GitHub API to get latest release
        api_url = (
            "https://api.github.com/repos/saigontechnology/AgentCrew/releases/latest"
        )
        response = requests.get(api_url, timeout=10)

        if response.status_code == 200:
            release_data = response.json()
            return release_data.get("tag_name", "").lstrip("v")

        # Fallback: get tags and find the latest
        tags_url = "https://api.github.com/repos/saigontechnology/AgentCrew/tags"
        response = requests.get(tags_url, timeout=10)

        if response.status_code == 200:
            tags_data = response.json()
            if tags_data:
                # Get the first (latest) tag
                latest_tag = tags_data[0].get("name", "").lstrip("v")
                return latest_tag

        return None
    except Exception:
        return None


def version_is_older(current: str, latest: str) -> bool:
    """
    Compare two semantic version strings to check if current is older than latest.

    Args:
        current: Current version string (e.g., "0.5.1")
        latest: Latest version string (e.g., "0.6.0")

    Returns:
        True if current version is older than latest version
    """
    try:
        # Clean version strings (remove 'v' prefix if present)
        current_clean = current.lstrip("v")
        latest_clean = latest.lstrip("v")

        # Split version strings into components
        current_parts = [int(x) for x in current_clean.split(".")]
        latest_parts = [int(x) for x in latest_clean.split(".")]

        # Pad shorter version with zeros for comparison
        max_length = max(len(current_parts), len(latest_parts))
        current_parts.extend([0] * (max_length - len(current_parts)))
        latest_parts.extend([0] * (max_length - len(latest_parts)))

        # Compare version components
        for current_part, latest_part in zip(current_parts, latest_parts):
            if current_part < latest_part:
                return True
            elif current_part > latest_part:
                return False

        # Versions are equal
        return False

    except (ValueError, AttributeError):
        # If version parsing fails, fall back to string comparison
        return current != latest


def run_update_command():
    """Run the appropriate update command based on the operating system"""
    try:
        system = platform.system().lower()

        if system == "linux" or system == "darwin":  # Darwin is macOS
            # Linux/macOS update command
            command = (
                "uv tool install --reinstall --force --python=3.12 agentcrew-ai@latest"
            )
            click.echo("🐧 Running Linux/macOS update command...")

        elif system == "windows":
            # Windows update command
            command = (
                "uv tool install --reinstall --force --python=3.12 agentcrew-ai@latest"
            )
            click.echo("🪟 Running Windows update command...")

        else:
            click.echo(f"❌ Unsupported operating system: {system}", err=True)
            return

        # Execute the update command
        result = subprocess.run(command, shell=True, capture_output=True, text=True)

        if result.returncode == 0:
            click.echo("✅ Update completed successfully!")
            click.echo("🔄 Please restart the application to use the new version.")
        else:
            click.echo("❌ Update failed!")
            if result.stderr:
                click.echo(f"Error: {result.stderr}")

    except Exception as e:
        click.echo(f"❌ Update execution failed: {str(e)}", err=True)


def setup_services(provider, memory_llm=None):
    # Initialize the model registry and service manager
    registry = ModelRegistry.get_instance()
    llm_manager = ServiceManager.get_instance()

    # Set the current model based on provider
    models = registry.get_models_by_provider(provider)
    if models:
        # Find default model for this provider
        default_model = next((m for m in models if m.default), models[0])
        registry.set_current_model(f"{default_model.provider}/{default_model.id}")

    # Get the LLM service from the manager
    llm_service = llm_manager.get_service(provider)

    try:
        config_manager = ConfigManagement()
        last_model = config_manager.get_last_used_model()
        last_provider = config_manager.get_last_used_provider()

        # Only restore if the last used provider matches current provider or if no specific provider was requested
        if last_model and last_provider:
            # Check if we should restore the last used model
            should_restore = False
            if provider == last_provider:
                # Same provider, definitely restore
                should_restore = True
            elif provider is None:
                # No specific provider requested, try to restore anyway
                should_restore = True

            last_model_class = registry.get_model(last_model)
            if should_restore and last_model_class:
                llm_service.model = last_model_class.id
    except Exception as e:
        # Don't fail startup if restoration fails
        click.echo(f"⚠️  Could not restore last used model: {e}")

    # Initialize services
    if memory_llm:
        memory_service = ChromaMemoryService(
            llm_service=llm_manager.initialize_standalone_service(memory_llm)
        )
    else:
        memory_service = ChromaMemoryService(
            llm_service=llm_manager.initialize_standalone_service(provider)
        )

    context_service = ContextPersistenceService()
    clipboard_service = ClipboardService()
    # Try to create search service if API key is available
    try:
        search_service = TavilySearchService()
    except Exception as e:
        click.echo(f"⚠️ Web search tools not available: {str(e)}")
        search_service = None

    # Initialize code analysis service
    try:
        code_analysis_service = CodeAnalysisService()
    except Exception as e:
        click.echo(f"⚠️ Code analysis tool not available: {str(e)}")
        code_analysis_service = None

    # Initialize image generation service
    try:
        # Use OpenAI by default if API key is available
        if os.getenv("OPENAI_API_KEY"):
            image_gen_service = ImageGenerationService()
        else:
            image_gen_service = None
            click.echo("⚠️ Image generation service not available: No API keys found.")
    except Exception as e:
        click.echo(f"⚠️ Image generation service not available: {str(e)}")
        image_gen_service = None

    # Clean up old memories (older than 1 month)
    try:
        removed_count = memory_service.cleanup_old_memories(months=1)
        if removed_count > 0:
            click.echo(f"🧹 Cleaned up {removed_count} old conversation memories")
    except Exception as e:
        click.echo(f"⚠️ Memory cleanup failed: {str(e)}")

    # Register all tools with their respective services
    services = {
        "llm": llm_service,
        "memory": memory_service,
        "clipboard": clipboard_service,
        "code_analysis": code_analysis_service,
        "web_search": search_service,
        "context_persistent": context_service,
        "image_generation": image_gen_service,
    }
    return services


def setup_agents(services, config_path, remoting_provider=None, model_id=None):
    """
    Set up the agent system with specialized agents.

    Args:
        services: Dictionary of services
    """
    # Get the singleton instance of agent manager
    agent_manager = AgentManager.get_instance()
    llm_manager = ServiceManager.get_instance()

    # Add agent_manager to services for tool registration
    services["agent_manager"] = agent_manager

    # Get the LLM service
    llm_service = services["llm"]

    # Create specialized agents
    if config_path:
        click.echo("Using command-line argument for agent configuration path.")
        os.environ["SW_AGENTS_CONFIG"] = config_path
    else:
        config_path = os.getenv("SW_AGENTS_CONFIG")
        if not config_path:
            config_path = "./agents.toml"
            # If config path doesn't exist, create a default one
        if not os.path.exists(config_path):
            click.echo(
                f"Agent configuration not found at {config_path}. Creating default configuration."
            )
            os.makedirs(os.path.dirname(config_path), exist_ok=True)

            default_config = """# Default SwissKnife Agent Configuration
[[agents]]
name = "default"
description = "Default assistant agent"
system_prompt = \"\"\"You are a helpful AI assistant. Always provide accurate, helpful, and ethical responses.
Current date: {current_date}
\"\"\"
tools = ["memory", "clipboard", "web_search", "code_analysis"]
"""

            with open(config_path, "w+", encoding="utf-8") as f:
                f.write(default_config)

            click.echo(f"Created default agent configuration at {config_path}")
    # Load agents from configuration
    agent_definitions = AgentManager.load_agents_from_config(config_path)
    first_agent_name = None
    for agent_def in agent_definitions:
        if agent_def.get("base_url", ""):
            try:
                agent = RemoteAgent(
                    agent_def["name"],
                    agent_def.get("base_url"),
                    headers=agent_def.get("headers", {}),
                )
            except Exception:
                print("Error: cannot connect to remote agent, skipping...")
                continue
        else:
            if not first_agent_name:
                first_agent_name = agent_def["name"]
            if remoting_provider:
                llm_service = llm_manager.initialize_standalone_service(
                    remoting_provider
                )
                if model_id:
                    llm_service.model = model_id
            agent = LocalAgent(
                name=agent_def["name"],
                description=agent_def["description"],
                llm_service=llm_service,
                services=services,
                tools=agent_def["tools"],
                temperature=agent_def.get("temperature", None),
            )
            agent.set_system_prompt(agent_def["system_prompt"])
            if remoting_provider:
                agent.set_custom_system_prompt(agent_manager.get_remote_system_prompt())
                agent.is_remoting_mode = True
                agent.activate()
        agent_manager.register_agent(agent)

    from AgentCrew.modules.mcpclient.tool import register as mcp_register

    mcp_register()

    if remoting_provider:
        from AgentCrew.modules.mcpclient import MCPSessionManager

        mcp_manager = MCPSessionManager.get_instance()
        mcp_manager.initialize_for_agent()
        return

    initial_agent_selected = False
    try:
        config_manager = ConfigManagement()
        last_agent = config_manager.get_last_used_agent()

        if last_agent and last_agent in agent_manager.agents:
            if agent_manager.select_agent(last_agent):
                initial_agent_selected = True
    except Exception as e:
        # Don't fail startup if restoration fails
        click.echo(f"⚠️  Could not restore last used agent: {e}")

    # Select the initial agent if specified and no agent was restored
    if not initial_agent_selected and first_agent_name:
        if not agent_manager.select_agent(first_agent_name):
            available_agents = ", ".join(agent_manager.agents.keys())
            click.echo(
                f"⚠️ Unknown agent: {first_agent_name}. Using default agent. Available agents: {available_agents}"
            )


def discover_and_register_tools(services=None):
    """
    Discover and register all tools

    DEPRECATED: This function is deprecated and will be removed in a future version.
    Use setup_agents() and register_agent_tools() instead.

    Args:
        services: Dictionary mapping service names to service instances
    """
    import warnings

    warnings.warn(
        "discover_and_register_tools is deprecated. Use setup_agents() and register_agent_tools() instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    if services is None:
        services = {}

    # List of tool modules and their corresponding service keys
    tool_modules = [
        ("modules.memory.tool", "memory"),
        ("modules.code_analysis.tool", "code_analysis"),
        ("modules.clipboard.tool", "clipboard"),
        ("modules.web_search.tool", "web_search"),
        ("modules.coding.tool", "aider"),
        ("modules.image_generation.tool", "image_generation"),
    ]

    for module_name, service_key in tool_modules:
        try:
            module = importlib.import_module(module_name)
            if hasattr(module, "register"):
                service_instance = services.get(service_key)
                module.register(service_instance)
                # print(f"✅ Registered tools from {module_name}")
        except ImportError as e:
            print(f"⚠️ Error importing tool module {module_name}: {e}")

    from AgentCrew.modules.mcpclient.tool import register as mcp_register

    mcp_register()


@cli.command()
@click.option(
    "--provider",
    type=click.Choice(
        ["claude", "groq", "openai", "google", "deepinfra", "github_copilot"]
    ),
    default=None,
    help="LLM provider to use (claude, groq, openai, google, github_copilot, or deepinfra)",
)
@click.option(
    "--agent-config", default=None, help="Path to the agent configuration file."
)
@click.option(
    "--mcp-config", default=None, help="Path to the mcp servers configuration file."
)
@click.option(
    "--memory-llm",
    type=click.Choice(
        ["claude", "groq", "openai", "google", "deepinfra", "github_copilot"]
    ),
    default=None,
    help="LLM Model use for analyzing and processing memory",
)
@click.option(
    "--console",
    is_flag=True,
    default=False,
    help="Use console interface instead of GUI",
)
def chat(provider, agent_config, mcp_config, memory_llm, console):
    """Start an interactive chat session with LLM"""
    try:
        load_api_keys_from_config()

        # Only check environment variables if provider wasn't explicitly specified
        if provider is None:
            # NEW: First try to restore last used provider
            try:
                config_manager = ConfigManagement()
                last_provider = config_manager.get_last_used_provider()
                if last_provider:
                    # Verify the provider is still available
                    if last_provider in [
                        "claude",
                        "groq",
                        "openai",
                        "google",
                        "deepinfra",
                        "github_copilot",
                    ]:
                        # Check if API key is available for this provider
                        api_key_map = {
                            "claude": "ANTHROPIC_API_KEY",
                            "google": "GEMINI_API_KEY",
                            "openai": "OPENAI_API_KEY",
                            "groq": "GROQ_API_KEY",
                            "deepinfra": "DEEPINFRA_API_KEY",
                            "github_copilot": "GITHUB_COPILOT_API_KEY",
                        }
                        if os.getenv(api_key_map.get(last_provider, "")):
                            provider = last_provider
                    else:
                        # Check if it's a custom provider
                        custom_providers = (
                            config_manager.read_custom_llm_providers_config()
                        )
                        if any(p["name"] == last_provider for p in custom_providers):
                            provider = last_provider
            except Exception as e:
                click.echo(f"⚠️  Could not restore last used provider: {e}")

            # Fall back to environment variable detection if no provider restored
            if provider is None:
                if os.getenv("GITHUB_COPILOT_API_KEY"):
                    provider = "github_copilot"
                elif os.getenv("ANTHROPIC_API_KEY"):
                    provider = "claude"
                elif os.getenv("GEMINI_API_KEY"):
                    provider = "google"
                elif os.getenv("OPENAI_API_KEY"):
                    provider = "openai"
                elif os.getenv("GROQ_API_KEY"):
                    provider = "groq"
                elif os.getenv("DEEPINFRA_API_KEY"):
                    provider = "deepinfra"
                else:
                    config = ConfigManagement()
                    custom_providers = config.read_custom_llm_providers_config()
                    if len(custom_providers) > 0:
                        # Use the first custom provider as default if no API keys found
                        provider = custom_providers[0]["name"]
                    else:
                        # Ask user to setup api key if nothing found
                        from AgentCrew.modules.gui.widgets.config_window import (
                            ConfigWindow,
                        )

                        app = QApplication(sys.argv)
                        config_window = ConfigWindow()
                        config_window.tab_widget.setCurrentIndex(3)  # Show Settings tab
                        config_window.show()
                        sys.exit(app.exec())
        services = setup_services(provider, memory_llm)

        if mcp_config:
            os.environ["MCP_CONFIG_PATH"] = mcp_config

        # Set up the agent system
        setup_agents(services, agent_config)

        # Create the message handler
        message_handler = MessageHandler(
            services["memory"], services["context_persistent"]
        )

        # Choose between GUI and console based on the --gui flag
        if not console:
            app = QApplication(sys.argv)
            chat_window = ChatWindow(message_handler)
            chat_window.show()
            sys.exit(app.exec())
        else:
            ui = ConsoleUI(message_handler)
            ui.start()
    except Exception as e:
        print(traceback.format_exc())
        click.echo(f"❌ Error: {str(e)}", err=True)
    finally:
        from AgentCrew.modules.mcpclient import MCPSessionManager

        MCPSessionManager.get_instance().cleanup()


@cli.command()
@click.option("--host", default="0.0.0.0", help="Host to bind the server to")
@click.option("--port", default=41241, help="Port to bind the server to")
@click.option("--base-url", default=None, help="Base URL for agent endpoints")
@click.option(
    "--provider",
    type=click.Choice(
        ["claude", "groq", "openai", "google", "github_copilot", "deepinfra"]
    ),
    default=None,
    help="LLM provider to use (claude, groq, openai, google, github_copilot or deepinfra)",
)
@click.option("--model-id", default=None, help="Model ID from provider")
@click.option("--agent-config", default=None, help="Path to agent configuration file")
@click.option("--api-key", default=None, help="API key for authentication (optional)")
@click.option(
    "--mcp-config", default=None, help="Path to the mcp servers configuration file."
)
@click.option(
    "--memory-llm",
    type=click.Choice(["claude", "groq", "openai", "google"]),
    default=None,
    help="LLM Model use for analyzing and processing memory",
)
def a2a_server(
    host,
    port,
    base_url,
    provider,
    model_id,
    agent_config,
    api_key,
    mcp_config,
    memory_llm,
):
    """Start an A2A server exposing all SwissKnife agents"""
    try:
        load_api_keys_from_config()

        # Set default base URL if not provided
        if not base_url:
            base_url = f"http://{host}:{port}"

        if provider is None:
            if os.getenv("GITHUB_COPILOT_API_KEY"):
                provider = "github_copilot"
            elif os.getenv("ANTHROPIC_API_KEY"):
                provider = "claude"
            elif os.getenv("GEMINI_API_KEY"):
                provider = "google"
            elif os.getenv("OPENAI_API_KEY"):
                provider = "openai"
            elif os.getenv("GROQ_API_KEY"):
                provider = "groq"
            elif os.getenv("DEEPINFRA_API_KEY"):
                provider = "deepinfra"
            else:
                raise ValueError(
                    "No LLM API key found. Please set either ANTHROPIC_API_KEY, GEMINI_API_KEY, OPENAI_API_KEY, GROQ_API_KEY, or DEEPINFRA_API_KEY"
                )

        services = setup_services(provider, memory_llm)
        if mcp_config:
            os.environ["MCP_CONFIG_PATH"] = mcp_config

        # Set up agents from configuration
        setup_agents(services, agent_config, provider, model_id)

        # Get agent manager
        agent_manager = AgentManager.get_instance()

        # Create and start server
        from AgentCrew.modules.a2a.server import A2AServer

        server = A2AServer(
            agent_manager=agent_manager,
            host=host,
            port=port,
            base_url=base_url,
            api_key=api_key,
        )

        click.echo(f"Starting A2A server on {host}:{port}")
        click.echo(f"Available agents: {', '.join(agent_manager.agents.keys())}")
        server.start()
    except Exception as e:
        print(traceback.format_exc())
        click.echo(f"❌ Error: {str(e)}", err=True)
    finally:
        from AgentCrew.modules.mcpclient import MCPSessionManager

        MCPSessionManager.get_instance().cleanup()


@cli.command()
def copilot_auth():
    """Authenticate with GitHub Copilot and save the API key to config"""
    try:
        click.echo("🔐 Starting GitHub Copilot authentication...")

        # Step 1: Request device code
        resp = requests.post(
            "https://github.com/login/device/code",
            headers={
                "accept": "application/json",
                "editor-version": "vscode/1.100.3",
                "editor-plugin-version": "GitHub.copilot/1.330.0",
                "content-type": "application/json",
                "user-agent": "GithubCopilot/1.330.0",
                "accept-encoding": "gzip,deflate,br",
            },
            data='{"client_id":"Iv1.b507a08c87ecfe98","scope":"read:user"}',
        )

        if resp.status_code != 200:
            click.echo(f"❌ Failed to get device code: {resp.status_code}", err=True)
            return

        # Parse the response json, isolating the device_code, user_code, and verification_uri
        resp_json = resp.json()
        device_code = resp_json.get("device_code")
        user_code = resp_json.get("user_code")
        verification_uri = resp_json.get("verification_uri")

        if not all([device_code, user_code, verification_uri]):
            click.echo("❌ Invalid response from GitHub", err=True)
            return

        # Print the user code and verification uri
        click.echo(f"📋 Please visit {verification_uri} and enter code: {user_code}")
        click.echo("⏳ Waiting for authentication...")

        # Step 2: Poll for access token
        while True:
            time.sleep(5)

            resp = requests.post(
                "https://github.com/login/oauth/access_token",
                headers={
                    "accept": "application/json",
                    "editor-version": "vscode/1.100.3",
                    "editor-plugin-version": "GitHub.copilot/1.330.0",
                    "content-type": "application/json",
                    "user-agent": "GithubCopilot/1.330.0",
                    "accept-encoding": "gzip,deflate,br",
                },
                data=f'{{"client_id":"Iv1.b507a08c87ecfe98","device_code":"{device_code}","grant_type":"urn:ietf:params:oauth:grant-type:device_code"}}',
            )

            # Parse the response json
            resp_json = resp.json()
            access_token = resp_json.get("access_token")
            error = resp_json.get("error")

            if access_token:
                click.echo("✅ Authentication successful!")
                break
            elif error == "authorization_pending":
                continue  # Keep polling
            elif error == "slow_down":
                time.sleep(5)  # Additional delay
                continue
            elif error == "expired_token":
                click.echo("❌ Authentication expired. Please try again.", err=True)
                return
            elif error == "access_denied":
                click.echo("❌ Authentication denied by user.", err=True)
                return
            else:
                click.echo(f"❌ Authentication error: {error}", err=True)
                return

        # Step 3: Save the token to config
        config_manager = ConfigManagement()
        global_config = config_manager.read_global_config_data()

        # Ensure api_keys section exists
        if "api_keys" not in global_config:
            global_config["api_keys"] = {}

        # Save the token
        global_config["api_keys"]["GITHUB_COPILOT_API_KEY"] = access_token
        config_manager.write_global_config_data(global_config)

        click.echo("💾 GitHub Copilot API key saved to config file!")
        click.echo("🚀 You can now use GitHub Copilot with --provider github_copilot")

    except ImportError:
        click.echo(
            "❌ Error: 'requests' package is required for authentication", err=True
        )
        click.echo("Install it with: pip install requests")
    except Exception as e:
        click.echo(f"❌ Authentication failed: {str(e)}", err=True)


if __name__ == "__main__":
    """Check for updates and update AgentCrew if a new version is available"""
    cli()



from typing import Dict, List, Any

from AgentCrew.modules.llm.base import BaseLLMService


class ConversationConsolidator:
    """
    A class for consolidating conversation messages to reduce token usage while preserving key information.
    Handles summarizing older messages and inserting consolidated message markers.
    """

    def __init__(self, llm_service: BaseLLMService):
        """
        Initialize the ConversationConsolidator.

        Args:
            message_handler: The MessageHandler instance to access messages and LLM services.
        """
        self.llm = llm_service

    async def consolidate(
        self, messages: List[Dict[str, Any]], preserve_count: int
    ) -> Dict[str, Any]:
        """
        Consolidate conversation messages, preserving the specified number of recent messages.

        Args:
            preserve_count: Number of most recent messages to preserve intact

        Returns:
            A dictionary with information about the consolidation (messages consolidated, tokens saved, etc.)
        """
        if preserve_count < 1:
            raise ValueError("preserve_count must be a positive integer")

        all_messages = messages

        if len(all_messages) <= preserve_count:
            return {
                "success": False,
                "reason": "Not enough messages to consolidate",
                "total_messages": len(all_messages),
                "preserve_count": preserve_count,
            }

        last_consolidated_idx = -1
        for i, msg in reversed(list(enumerate(all_messages))):
            if msg.get("role") == "consolidated":
                last_consolidated_idx = i
                break

        if all_messages[-preserve_count].get("role", "") == "tool":
            for i, msg in reversed(list(enumerate(all_messages[:-preserve_count]))):
                if msg.get("role") == "assistant":
                    preserve_count += len(all_messages[:-preserve_count]) - i
                    break

        if last_consolidated_idx >= 0:
            to_consolidate = (
                all_messages[last_consolidated_idx + 1 : -preserve_count]
                if preserve_count > 0
                else all_messages[last_consolidated_idx + 1 :]
            )
            to_preserve = all_messages[-preserve_count:] if preserve_count > 0 else []

        else:
            to_consolidate = (
                all_messages[:-preserve_count] if preserve_count > 0 else all_messages
            )
            to_preserve = all_messages[-preserve_count:] if preserve_count > 0 else []

        if not to_consolidate:
            return {
                "success": False,
                "reason": "No messages to consolidate after preserving recent messages",
                "total_messages": len(all_messages),
                "preserve_count": preserve_count,
            }

        original_tokens = self.estimate_token_count(to_consolidate)

        summary = await self.generate_summary(to_consolidate)

        consolidated_msg = {
            "role": "consolidated",
            "content": [{"type": "text", "text": summary}],
            "metadata": {
                "messages_consolidated": len(to_consolidate),
                "original_token_count": original_tokens,
                "consolidated_token_count": self.estimate_token_count(
                    [{"content": summary}]
                ),
            },
        }

        consolidated_tokens = self.estimate_token_count([consolidated_msg])

        messages.insert(-preserve_count, consolidated_msg)

        result = {
            "success": True,
            "messages_consolidated": len(to_consolidate),
            "messages_preserved": len(to_preserve),
            "original_token_count": original_tokens,
            "consolidated_token_count": consolidated_tokens,
            "summary": summary,
        }

        return result

    async def generate_summary(self, messages: List[Dict[str, Any]]) -> str:
        """
        Generate a summary of the given messages using the current agent's LLM.

        Args:
            messages: The messages to summarize

        Returns:
            A string summary of the messages
        """

        formatted_conversation = self.format_conversation_for_summary(messages)

        summary_prompt = self.create_summary_prompt(formatted_conversation)

        summary = await self.llm.process_message(summary_prompt)

        return summary

    def format_conversation_for_summary(self, messages: List[Dict[str, Any]]) -> str:
        """
        Format conversation messages for summarization.

        Args:
            messages: The messages to format

        Returns:
            Formatted conversation as a string
        """
        formatted_parts = []

        for msg in messages:
            role = msg.get("role", "")
            agent_name = msg.get("agent", "unknown")

            if role == "user":
                prefix = "USER: "
            elif role == "assistant":
                prefix = f"{agent_name.upper()}: "
            elif role == "tool":
                prefix = f"TOOL ({msg.get('name', 'unknown')}): "
            elif role == "consolidated":
                # If we're summarizing a section that already has a consolidated message,
                # include it directly to preserve that context
                prefix = "PREVIOUS SUMMARY: "
            else:
                continue

            content = msg.get("content", "")

            if isinstance(content, str):
                text_content = content
            elif isinstance(content, list):
                text_parts = []
                for part in content:
                    if isinstance(part, dict):
                        if part.get("type") == "text":
                            text_parts.append(part.get("text", ""))
                        elif part.get("type") == "image_url":
                            text_parts.append("[IMAGE]")
                text_content = " ".join(text_parts)
            else:
                text_content = str(content)

            formatted_parts.append(f"{prefix}{text_content}")

        return "\n\n".join(formatted_parts)

    def create_summary_prompt(self, conversation: str) -> str:
        """
        Create a prompt for the LLM to generate a conversation summary.

        Args:
            conversation: The formatted conversation text

        Returns:
            A prompt string
        """
        return (
            """You are an expert conversation summarizer tasked with creating a concise but comprehensive summary of a conversation between a user and AI assistants.

IMPORTANT GUIDELINES:
1. Preserve critical information in these categories:
   - Key questions and their complete answers
   - Important decisions and conclusions
   - Technical details, facts, code snippets, and data
   - Project and key files location and project structure
   - Problems identified and their solutions
   - Commitments or agreements made
   
2. Structure your summary effectively:
   - Begin with a clear overview statement: "This conversation covers [main topics]"
   - Organize by main topics rather than chronologically when appropriate
   - Use bullet points or paragraphs as needed for clarity
   - Format code snippets with markdown code blocks (```language\ncode\n```)
   
3. Focus on content quality:
   - Be factual and objective - avoid adding information not in the original
   - Include specific details that would be needed to continue the conversation
   - Preserve the tone and style of technical information
   - Maintain technical accuracy and precision
   
4. Optimize for efficiency:
   - Aim for 15-20% of the original length
   - Eliminate redundancies and tangential discussions
   - Focus on substance over conversational patterns
   - Be comprehensive yet concise

The conversation to summarize is delimited between triple quotes:
"""
            + f'"""\n{conversation}\n"""'
        )

    async def unconsolidate(self, messages: List[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Remove the last consolidated message and restore original messages.

        Args:
            messages: The current message list containing consolidated messages

        Returns:
            A dictionary with information about the unconsolidation
        """
        # Find the last consolidated message
        last_consolidated_idx = -1
        consolidated_message = None

        for i, msg in reversed(list(enumerate(messages))):
            if msg.get("role") == "consolidated":
                last_consolidated_idx = i
                consolidated_message = msg
                break

        if last_consolidated_idx == -1 or consolidated_message is None:
            return {
                "success": False,
                "reason": "No consolidated message found to unconsolidate",
                "total_messages": len(messages),
            }

        metadata = consolidated_message.get("metadata", {})
        messages_consolidated = metadata.get("messages_consolidated", 0)

        messages.pop(last_consolidated_idx)

        return {
            "success": True,
            "messages_restored": messages_consolidated,
            "consolidated_message_removed": True,
            "restoration_note": "Consolidated message removed. Original messages would need to be restored from backup if available.",
        }

    def has_consolidated_message(self, messages: List[Dict[str, Any]]) -> bool:
        """
        Check if the message list contains any consolidated messages.

        Args:
            messages: The message list to check

        Returns:
            True if consolidated messages exist, False otherwise
        """
        return any(msg.get("role") == "consolidated" for msg in messages)

    def estimate_token_count(self, messages: List[Dict[str, Any]]) -> int:
        """
        Estimate the token count for the given messages.

        Args:
            messages: The messages to count tokens for

        Returns:
            An estimated token count
        """
        # Simple approximation - 1 token ~= 4 characters
        total_chars = 0

        for msg in messages:
            content = msg.get("content", "")

            if isinstance(content, str):
                total_chars += len(content)
            elif isinstance(content, list):
                for part in content:
                    if isinstance(part, dict) and part.get("type") == "text":
                        total_chars += len(part.get("text", ""))

        return total_chars // 4  # Rough approximation



"""
User input handling for console UI.
Manages user input threads, key bindings, and prompt sessions.
"""

import asyncio
import sys
import time
import threading
import queue
from threading import Thread, Event
from prompt_toolkit import PromptSession
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.formatted_text import HTML
from rich.console import Console
from rich.text import Text

from AgentCrew.modules import logger
from AgentCrew.modules.chat import MessageHandler
from AgentCrew.modules.clipboard.service import ClipboardService
from .completers import ChatCompleter
from .display_handlers import DisplayHandlers
from .constants import (
    RICH_STYLE_YELLOW,
    RICH_STYLE_YELLOW_BOLD,
    RICH_STYLE_RED,
    RICH_STYLE_BLUE,
)


class InputHandler:
    """Handles user input in a separate thread and manages key bindings."""

    def __init__(
        self,
        console: Console,
        message_handler: MessageHandler,
        display_handlers: DisplayHandlers,
    ):
        """Initialize the input handler."""
        self.console = console
        self.message_handler = message_handler
        self.display_handlers = display_handlers
        self.clipboard_service = ClipboardService()

        # Threading for user input
        self._input_queue = queue.Queue()
        self._input_thread = None
        self._input_stop_event = Event()
        self._current_prompt_session = None
        self._last_ctrl_c_time = 0
        self.is_message_processing = False

        # Set up key bindings
        self.kb = self._setup_key_bindings()

    def _setup_key_bindings(self):
        """Set up key bindings for multiline input."""
        kb = KeyBindings()

        @kb.add(Keys.ControlS)
        @kb.add("escape", "enter")
        def _(event):
            """Submit on Ctrl+S."""
            if event.current_buffer.text.strip() and not self.is_message_processing:
                event.current_buffer.validate_and_handle()

        @kb.add(Keys.Enter)
        def _(event):
            """Insert newline on Enter."""
            event.current_buffer.insert_text("\n")

        @kb.add("escape", "c")  # Alt+C
        def _(event):
            """Copy latest assistant response to clipboard."""
            # This will be handled by the main console UI
            pass

        @kb.add(Keys.ControlV)
        def _(event):
            """Handle Ctrl+V with image/binary detection."""
            try:
                # Check if clipboard contains image or binary content
                paste_result = self.clipboard_service.read_and_process_paste()

                if paste_result["success"]:
                    content_type = paste_result.get("type")

                    if content_type == "file_command":
                        # Insert the file command
                        file_command = paste_result["content"]

                        # Insert the file command into the current buffer
                        event.current_buffer.insert_text(file_command)
                        event.current_buffer.validate_and_handle()

                        return

                # For regular text content, use default paste behavior
                event.current_buffer.paste_clipboard_data(
                    event.app.clipboard.get_data()
                )

            except Exception:
                # Fall back to default paste behavior if anything goes wrong
                try:
                    event.current_buffer.paste_clipboard_data(
                        event.app.clipboard.get_data()
                    )
                except Exception:
                    pass  # Ignore if even default paste fails

        @kb.add(Keys.ControlC)
        def _(event):
            """Handle Ctrl+C with confirmation for exit."""
            current_time = time.time()
            if (
                hasattr(self, "_last_ctrl_c_time")
                and current_time - self._last_ctrl_c_time <= 1
            ):
                self.console.print(
                    Text("\n🎮 Confirmed exit. Goodbye!", style=RICH_STYLE_YELLOW_BOLD)
                )
                # Don't try to join from within the same thread - just exit
                event.app.exit("__EXIT__")
            else:
                self._last_ctrl_c_time = current_time
                if (
                    hasattr(self.message_handler, "stream_generator")
                    and self.message_handler.stream_generator
                ):
                    try:
                        asyncio.run(self.message_handler.stream_generator.aclose())
                    except RuntimeError as e:
                        logger.warning(f"Error closing stream generator: {e}")
                    except Exception as e:
                        logger.warning(f"Exception closing stream generator: {e}")
                    finally:
                        self.message_handler.stop_streaming = True
                        self.message_handler.stream_generator = None
                else:
                    self.console.print(
                        Text(
                            "\nPress Ctrl+C again within 1 second to exit.",
                            style=RICH_STYLE_YELLOW,
                        )
                    )
                    self.display_handlers.print_prompt_prefix(
                        self.message_handler.agent.name,
                        self.message_handler.agent.get_model(),
                    )
                    prompt = Text("👤 YOU: ", style=RICH_STYLE_BLUE)
                    self.console.print(prompt, end="")
                    time.sleep(0.2)
                    self.clear_buffer()

        @kb.add(Keys.Up)
        def _(event):
            """Navigate to previous history entry."""
            buffer = event.current_buffer
            document = buffer.document

            # Check if cursor is at the first line's start
            cursor_position = document.cursor_position
            if document.cursor_position_row == 0 and cursor_position <= len(
                document.current_line
            ):
                # Get previous history entry
                prev_entry = self.message_handler.history_manager.get_previous()
                if prev_entry is not None:
                    # Replace current text with history entry
                    buffer.text = prev_entry
                    # Move cursor to end of text
                    buffer.cursor_position = len(prev_entry)
            else:
                # Regular up arrow behavior - move cursor up
                buffer.cursor_up()

        @kb.add(Keys.Backspace)
        def _(event):
            if not event.current_buffer.text:
                prompt = Text("👤 YOU: ", style=RICH_STYLE_BLUE)
                self.console.print("", end="\r")
                self.console.print(prompt, end="")
            else:
                event.current_buffer.delete_before_cursor()

        @kb.add(Keys.Down)
        def _(event):
            """Navigate to next history entry if cursor is at last line."""
            buffer = event.current_buffer
            document = buffer.document

            # Check if cursor is at the last line
            if document.cursor_position_row == document.line_count - 1:
                # Get next history entry
                next_entry = self.message_handler.history_manager.get_next()
                if next_entry is not None:
                    # Replace current text with history entry
                    buffer.text = next_entry
                    # Move cursor to end of text
                    buffer.cursor_position = len(next_entry)
            else:
                # Regular down arrow behavior - move cursor down
                buffer.cursor_down()

        return kb

    def clear_buffer(self):
        if self._current_prompt_session:
            self._current_prompt_session.app.current_buffer.reset()
            self._current_prompt_session.message = HTML("<ansiblue>👤 YOU:</ansiblue> ")
            self._current_prompt_session.app.invalidate()

    def _input_thread_worker(self):
        """Worker thread for handling user input."""
        while not self._input_stop_event.is_set():
            try:
                session = PromptSession(
                    key_bindings=self.kb,
                    completer=ChatCompleter(self.message_handler),
                    refresh_interval=0.3,
                )
                self._current_prompt_session = session

                # Create a dynamic prompt that includes agent and model info using HTML formatting
                prompt_text = (
                    HTML("<ansiblue>👤 YOU:</ansiblue> ")
                    if not self.is_message_processing
                    else ""
                )
                user_input = session.prompt(prompt_text)

                # Reset history position after submission
                self.message_handler.history_manager.reset_position()

                # Put the input in the queue
                self._input_queue.put(user_input)
                time.sleep(0.4)  # Allow time for input processing

            except KeyboardInterrupt:
                # Handle Ctrl+C in input thread
                current_time = time.time()
                if (
                    hasattr(self, "_last_ctrl_c_time")
                    and current_time - self._last_ctrl_c_time < 2
                ):
                    self._input_queue.put("__EXIT__")
                    break
                else:
                    self._last_ctrl_c_time = current_time
                    self._input_queue.put("__INTERRUPT__")
                    continue
            except Exception as e:
                self._input_queue.put(f"__ERROR__:{str(e)}")
                break

    def _start_input_thread(self):
        """Start the input thread if not already running."""
        if self._input_thread is None or not self._input_thread.is_alive():
            self._input_stop_event.clear()
            self._input_thread = Thread(target=self._input_thread_worker, daemon=True)
            self._input_thread.start()

    def _stop_input_thread(self):
        """Stop the input thread cleanly."""
        if self._input_thread and self._input_thread.is_alive():
            # Don't try to join if we're in the same thread
            if threading.current_thread() == self._input_thread:
                # We're in the input thread, just set the stop event
                self._input_stop_event.set()
                return

            self._input_stop_event.set()
            if self._current_prompt_session:
                # Try to interrupt the current prompt session
                try:
                    if (
                        hasattr(self._current_prompt_session, "app")
                        and self._current_prompt_session.app
                    ):
                        self._current_prompt_session.app.exit()
                except Exception:
                    pass
            self._input_thread.join(timeout=1.0)

    def get_user_input(self):
        """
        Get multiline input from the user with support for command history.
        Now runs in a separate thread to allow events to display during input.

        Returns:
            The user input as a string.
        """
        # Start input thread if not already running
        if self._input_thread is None or not self._input_thread.is_alive():
            title = Text(f"\n[{self.message_handler.agent.name}", style=RICH_STYLE_RED)
            title.append(":")
            title.append(
                f"{self.message_handler.agent.get_model()}]",
                style=RICH_STYLE_BLUE,
            )
            title.append(
                "\n(Press Enter for new line, Ctrl+S/Alt+Enter to submit, Up/Down for history, Ctrl+V to paste)\n",
                style=RICH_STYLE_YELLOW,
            )
            self.console.print(title)
            self.display_handlers.display_added_files()
            self._start_input_thread()
        else:
            self.display_handlers.print_prompt_prefix(
                self.message_handler.agent.name, self.message_handler.agent.get_model()
            )
            self.clear_buffer()

        # Wait for input while allowing events to be processed
        while True:
            try:
                # Check for input with a short timeout to allow event processing
                user_input = self._input_queue.get(timeout=0.2)

                # Add None check here
                if user_input is None:
                    continue

                if user_input == "__EXIT__":
                    self.console.print(
                        Text(
                            "\n🎮 Confirmed exit. Goodbye!",
                            style=RICH_STYLE_YELLOW_BOLD,
                        )
                    )
                    self._stop_input_thread()
                    sys.exit(0)
                elif user_input == "__INTERRUPT__":
                    self.console.print(
                        Text(
                            "\n🎮 Chat interrupted. Press Ctrl+C again within 1 second to exit.",
                            style=RICH_STYLE_YELLOW_BOLD,
                        )
                    )
                    return ""
                elif user_input.startswith("__ERROR__:"):
                    error_msg = user_input[10:]  # Remove "__ERROR__:" prefix
                    self.console.print(
                        Text(f"\nInput error: {error_msg}", style=RICH_STYLE_RED)
                    )
                    return ""
                else:
                    self.display_handlers.display_divider()
                    return user_input

            except queue.Empty:
                # No input yet, continue waiting
                continue
            except KeyboardInterrupt:
                # Handle KeyboardInterrupt from the prompt session exit
                self.console.print(
                    Text(
                        "\n🎮 Confirmed exit. Goodbye!",
                        style=RICH_STYLE_YELLOW_BOLD,
                    )
                )
                sys.exit(0)

    def stop(self):
        """Stop the input handler and clean up."""
        self._stop_input_thread()



from .console_ui import ConsoleUI
from .display_handlers import DisplayHandlers
from .tool_display import ToolDisplayHandlers
from .input_handler import InputHandler
from .ui_effects import UIEffects
from .confirmation_handler import ConfirmationHandler
from .conversation_handler import ConversationHandler

__all__ = [
    "ConsoleUI",
    "DisplayHandlers",
    "ToolDisplayHandlers",
    "InputHandler",
    "UIEffects",
    "ConfirmationHandler",
    "ConversationHandler",
]



import os
from typing import Dict, Any, List
from dotenv import load_dotenv
from tavily import TavilyClient


class TavilySearchService:
    """Service for interacting with the Tavily Search API using the official SDK."""

    def __init__(self):
        """Initialize the Tavily search service with API key from environment."""
        load_dotenv()
        self.api_key = os.getenv("TAVILY_API_KEY")
        if not self.api_key:
            raise ValueError("TAVILY_API_KEY not found in environment variables")

        self.client = TavilyClient(api_key=self.api_key)

    def search(
        self,
        query: str,
        search_depth: str = "basic",
        topic: str = "general",
        include_domains: List[str] | None = None,
        exclude_domains: List[str] | None = None,
        max_results: int = 5,
    ) -> Dict[str, Any]:
        """
        Perform a web search using Tavily API.

        Args:
            query: The search query
            search_depth: 'basic' or 'advanced' search depth
            include_domains: List of domains to include in search
            exclude_domains: List of domains to exclude from search
            max_results: Maximum number of results to return

        Returns:
            Dict containing search results
        """
        try:
            params = {
                "query": query,
                "search_depth": search_depth,
                "max_results": max_results,
                "include_answer": search_depth,
                "topic": topic,
            }

            if include_domains:
                params["include_domains"] = include_domains

            if exclude_domains:
                params["exclude_domains"] = exclude_domains

            return self.client.search(**params)
        except Exception as e:
            print(f"❌ Search error: {str(e)}")
            return {"error": str(e)}

    def extract(self, url: str) -> Dict[str, Any]:
        """
        Extract content from a specific URL using Tavily API.

        Args:
            url: The URL to extract content from

        Returns:
            Dict containing the extracted content
        """
        try:
            return self.client.extract(url)
        except Exception as e:
            print(f"❌ Extract error: {str(e)}")
            return {"error": str(e)}

    def format_search_results(self, results: Dict[str, Any]) -> str:
        """Format search results into a readable string."""
        if "error" in results:
            return f"Search error: {results['error']}"
        formatted_text = ""

        if "answer" in results:
            formatted_text += f"**Query's Summary**: {results['answer']}  \n\n"

        formatted_text += "**Search Results**:  \n\n"

        if "results" in results:
            for i, result in enumerate(results["results"], 1):
                formatted_text += f"{i}. {result.get('title', 'No title')} (Matching score: {result.get('score', 'Unknown')}) \n"
                formatted_text += f"   URL: {result.get('url', 'No URL')}  \n"
                formatted_text += f"   {result.get('content', 'No content')}  \n\n"
        else:
            formatted_text += "No results found."

        return formatted_text

    def format_extract_results(self, results: Dict[str, Any]) -> str:
        """Format extract results into a readable string."""

        if "failed_results" in results and results["failed_results"]:
            result = results["failed_results"][0]
            return f"Extract failed: {result.get('error', 'Unknown error')}"

        if "results" in results and results["results"]:
            result = results["results"][0]
            url = result.get("url", "Unknown URL")
            content = result.get("raw_content", "No content available")
            # if self.llm:
            #     content = self.llm.summarize_content(content)
            return f"Extracted content from {url}:\n\n{content}"
        else:
            return "No content could be extracted."



from .message_handlers import MessageEventHandler
from .tool_handlers import ToolEventHandler
from .keyboard_handler import KeyboardHandler
from .menu_components import MenuBuilder
from .chat_components import ChatComponents
from .ui_state_manager import UIStateManager
from .input_components import InputComponents
from .conversation_components import ConversationComponents
from .command_handler import CommandHandler

__all__ = [
    "MessageEventHandler",
    "ToolEventHandler",
    "KeyboardHandler",
    "MenuBuilder",
    "ChatComponents",
    "UIStateManager",
    "InputComponents",
    "ConversationComponents",
    "CommandHandler",
]



import json
import os
import uuid
import datetime
from typing import Dict, Any, List, Optional
from AgentCrew.modules import logger


class ContextPersistenceService:
    """
    Manages persistence for user context (summary + rankings) and conversation
    histories for a single-user application, using JSON files.
    Handles nested structure for key_facts_entities.
    Persistence directory is determined by PERSISTENCE_DIR environment variable,
    defaulting to the current directory.
    Uses print for output and raises exceptions on critical errors.
    """

    CONVERSATIONS_SUBDIR = "conversations"
    ADAPTIVE_BEHAVIORS_FILE = "adaptive.json"

    def __init__(self, persistence_dir_override: Optional[str] = None):
        """
        Initializes the service, setting up paths and ensuring directories exist.

        The base directory is determined in the following order:
        1. `persistence_dir_override` argument (if provided).
        2. `PERSISTENCE_DIR` environment variable (if set and not empty).
        3. Current working directory (`.`) as the final default.

        Args:
            persistence_dir_override: Optional explicit path to the persistence directory,
                                      bypassing environment variable lookup.

        Raises:
            OSError: If the persistence directories cannot be created.
        """
        # Removed: self.logger initialization

        if persistence_dir_override:
            persistence_dir = persistence_dir_override
            # print(
            #     f"INFO: Using provided persistence directory override: {persistence_dir}"
            # )
        else:
            env_dir = os.getenv("PERSISTENCE_DIR")
            if env_dir:
                persistence_dir = env_dir
                # print(
                #     f"INFO: Using persistence directory from PERSISTENCE_DIR environment variable: {persistence_dir}"
                # )
            else:
                persistence_dir = "./persistents"  # Default to current directory
                # print(
                #     "INFO: PERSISTENCE_DIR environment variable not set. Defaulting persistence directory to current directory ('.')"
                # )

        # Expand user path (~) if present, and get absolute path for clarity
        self.base_dir = os.path.abspath(os.path.expanduser(persistence_dir))
        self.conversations_dir = os.path.join(self.base_dir, self.CONVERSATIONS_SUBDIR)
        self.adaptive_behaviors_file_path = os.path.join(
            self.base_dir, self.ADAPTIVE_BEHAVIORS_FILE
        )

        # _ensure_dir already raises OSError on failure
        self._ensure_dir(self.base_dir)
        self._ensure_dir(self.conversations_dir)
        logger.info(
            f"INFO: Persistence service initialized. Absolute base directory: {self.base_dir}"
        )

    def _ensure_dir(self, dir_path: str):
        """Ensures a directory exists, creating it if necessary."""
        try:
            os.makedirs(dir_path, exist_ok=True)
        except OSError as e:
            # Removed: self.logger.error(...)
            logger.error(f"ERROR: Failed to create directory {dir_path}: {e}")
            raise  # Re-raise after printing

    def _read_json_file(self, file_path: str, default_value: Any = None) -> Any:
        """
        Safely reads a JSON file. Returns default value on expected errors.

        Args:
            file_path: Path to the JSON file.
            default_value: Value to return if the file doesn't exist or is invalid.

        Returns:
            Parsed JSON content or the default value.
        """
        if not os.path.exists(file_path):
            return default_value
        try:
            with open(file_path, "r", encoding="utf-8") as f:
                content = f.read()
                if not content:
                    # Treat empty file same as invalid JSON for consistency
                    logger.warning(
                        f"WARNING: File {file_path} is empty. Returning default."
                    )
                    return default_value
                return json.loads(content)
        except (json.JSONDecodeError, IOError, UnicodeDecodeError) as e:
            # Removed: self.logger.warning(...)
            logger.warning(
                f"WARNING: Could not read or parse {file_path}: {e}. Returning default."
            )
            return default_value
        except Exception as e:
            # Catch unexpected errors during read/parse
            logger.error(f"ERROR: Unexpected error reading {file_path}: {e}")
            # Decide if unexpected errors should raise or return default.
            # Returning default might hide issues, raising might be better.
            # Let's raise for unexpected errors.
            raise

    def _write_json_file(self, file_path: str, data: Any):
        """
        Safely writes data to a JSON file. Raises exceptions on failure.

        Args:
            file_path: Path to the JSON file.
            data: Python object to serialize and write.

        Raises:
            IOError: If writing to the file fails.
            TypeError: If the data cannot be serialized to JSON.
            OSError: If the directory cannot be created.
        """
        try:
            # Ensure directory exists before writing (raises OSError on failure)
            self._ensure_dir(os.path.dirname(file_path))
            with open(file_path, "w", encoding="utf-8") as f:
                json.dump(data, f, indent=2, ensure_ascii=False)
        except (IOError, TypeError, OSError) as e:
            logger.error(f"ERROR: Could not write to {file_path}: {e}")
            raise  # Re-raise the caught exception
        except Exception as e:
            # Catch unexpected errors during write/dump
            logger.error(f"ERROR: Unexpected error writing {file_path}: {e}")
            raise

    # --- Conversation History Management ---

    def start_conversation(self) -> str:
        """
        Generates a unique conversation ID. Does not create a file immediately.

        Returns:
            The unique conversation ID (UUID string).
        """
        conversation_id = str(uuid.uuid4())
        # Removed file creation: File will be created on first append.
        # file_path = os.path.join(self.conversations_dir, f"{conversation_id}.json")
        # self._write_json_file(file_path, []) # REMOVED
        # print(f"INFO: Generated new conversation ID: {conversation_id}")
        return conversation_id

    def delete_conversation(self, conversation_id: str) -> bool:
        """
        Deletes a conversation JSON file from the filesystem.

        Args:
            conversation_id: The ID of the conversation to delete.

        Returns:
            True if the file was deleted or did not exist, False on error.
        """
        file_path = os.path.join(self.conversations_dir, f"{conversation_id}.json")
        try:
            if os.path.exists(file_path):
                os.remove(file_path)
                logger.info(f"INFO: Deleted conversation file: {file_path}")
            else:
                logger.info(
                    f"INFO: Conversation file not found (already deleted?): {file_path}"
                )
            return True
        except OSError as e:
            logger.error(f"ERROR: Failed to delete conversation file {file_path}: {e}")
            return False
        except Exception as e:
            logger.error(
                f"ERROR: Unexpected error deleting conversation file {file_path}: {e}"
            )
            return False

    def append_conversation_messages(
        self, conversation_id: str, new_messages: List[Dict[str, Any]], force=False
    ):
        """
        Appends a list of new message dictionaries to a conversation history file.

        Args:
            conversation_id: The ID of the conversation to update.
            new_messages: The list of new message dictionaries to append.
                          Typically contains a user message and an assistant message.

        Raises:
            ValueError: If new_messages format is invalid.
            IOError, TypeError, OSError: If reading or writing the conversation file fails.
        """
        if not isinstance(new_messages, list) or not all(
            isinstance(msg, dict) for msg in new_messages
        ):
            raise ValueError(
                f"Invalid new_messages format for {conversation_id} (must be a list of dicts). Aborting append."
            )

        if not new_messages and not force:
            # print(
            #     f"INFO: No new messages provided for {conversation_id}. Nothing to append."
            # )
            return  # Nothing to do

        file_path = os.path.join(self.conversations_dir, f"{conversation_id}.json")

        history = []  # Initialize history as empty list
        if os.path.exists(file_path):
            # File exists, read its content
            history = self._read_json_file(file_path, default_value=[])
            if not isinstance(history, list):
                logger.warning(
                    f"WARNING: Conversation file {file_path} was not a list. Resetting history before append."
                )
                history = []
        # else: File doesn't exist, history remains [], file will be created by _write_json_file

        if force:
            history = new_messages
        else:
            # Append the new messages
            history.extend(new_messages)

        self._write_json_file(file_path, history)
        # print(
        #     f"INFO: Appended {len(new_messages)} message(s) to conversation: {conversation_id}"
        # )

    def get_conversation_history(
        self, conversation_id: str
    ) -> List[Dict[str, Any]] | None:
        """
        Loads and returns the message list for a specific conversation.

        Args:
            conversation_id: The ID of the conversation to retrieve.

        Returns:
            A list of message dictionaries, or None if the conversation file
            doesn't exist or is invalid.
        """
        file_path = os.path.join(self.conversations_dir, f"{conversation_id}.json")
        history = self._read_json_file(file_path, default_value=None)

        if history is None or not isinstance(history, list):
            logger.warning(
                f"WARNING: Conversation history for {conversation_id} not found or invalid."
            )
            return None

        return history

    def list_conversations(self) -> List[Dict[str, Any]]:
        """
        Scans the conversations directory and returns metadata for available conversations.

        Returns:
            A list of dictionaries, each containing 'id', 'timestamp' (of last modification),
            and 'preview' (first few words of the first user message).
            Sorted by timestamp descending (most recent first).

        Raises:
            OSError: If the conversations directory cannot be listed.
        """
        conversations = []
        try:
            # listdir raises OSError if the directory is invalid
            filenames = os.listdir(self.conversations_dir)
            for filename in filenames:
                if filename.endswith(".json"):
                    conversation_id = filename[:-5]  # Remove .json extension
                    file_path = os.path.join(self.conversations_dir, filename)
                    try:
                        # getmtime raises OSError if file not found or inaccessible
                        mtime = os.path.getmtime(file_path)
                        timestamp = datetime.datetime.fromtimestamp(mtime).isoformat()

                        # _read_json_file handles its own errors internally
                        history = self._read_json_file(file_path, default_value=[])
                        preview = "Empty Conversation"
                        if isinstance(history, list) and len(history) > 0:
                            user_msgs = (
                                msg
                                for msg in history
                                if isinstance(msg, dict) and msg.get("role") == "user"
                            )
                            while True:
                                first_user_msg = next(
                                    user_msgs,
                                    None,
                                )
                                if first_user_msg:
                                    content = first_user_msg.get("content", "")
                                    if isinstance(content, str) and content:
                                        preview = (
                                            (content[:50] + "...")
                                            if len(content) > 50
                                            else content
                                        )
                                    elif isinstance(content, list):
                                        first_text_block = next(
                                            (
                                                block.get("text", "")
                                                for block in content
                                                if isinstance(block, dict)
                                                and block.get("type") == "text"
                                            ),
                                            "",
                                        )
                                        if first_text_block:
                                            preview = (
                                                (first_text_block[:50] + "...")
                                                if len(first_text_block) > 50
                                                else first_text_block
                                            )
                                        else:
                                            preview = "[Image/Tool Data]"
                                    else:
                                        preview = "[Non-text Content]"
                                else:
                                    preview = "[No User Message Found]"

                                if not preview.startswith(
                                    "Memories related to the user request:"
                                ) and not preview.startswith("Content of "):
                                    break

                        conversations.append(
                            {
                                "id": conversation_id,
                                "timestamp": timestamp,
                                "preview": preview,
                            }
                        )
                    except OSError as e:
                        # Log specific file access errors but continue listing others
                        logger.warning(
                            f"WARNING: Could not access metadata for {filename}: {e}"
                        )
                    except (
                        Exception
                    ) as e:  # Catch other potential errors during preview generation
                        logger.warning(
                            f"WARNING: Error processing {filename} for listing: {e}"
                        )

            # Sort by timestamp descending (most recent first)
            conversations.sort(key=lambda x: x["timestamp"], reverse=True)

        except FileNotFoundError:
            # This case might be less likely now due to __init__ checks, but keep for robustness
            logger.warning(
                f"WARNING: Conversations directory not found during listing: {self.conversations_dir}"
            )
        except OSError as e:
            # Raise error if listing the directory itself fails
            logger.warning(
                f"ERROR: Could not list conversations directory {self.conversations_dir}: {e}"
            )
            raise

        return conversations

    # --- Adaptive Behavior Management ---

    def get_adaptive_behaviors(self, agent_name: str) -> Dict[str, str]:
        """
        Retrieves all adaptive behaviors for a specific agent.

        Args:
            agent_name: The name of the agent.

        Returns:
            Dictionary of behavior ID to behavior description mappings.
        """
        adaptive_data = self._read_json_file(
            self.adaptive_behaviors_file_path, default_value={}
        )

        if not isinstance(adaptive_data, dict):
            logger.warning(
                "WARNING: Adaptive behaviors file was not a dictionary. Resetting."
            )
            return {}

        return adaptive_data.get(agent_name, {})

    def store_adaptive_behavior(
        self, agent_name: str, behavior_id: str, behavior: str
    ) -> bool:
        """
        Stores or updates an adaptive behavior for a specific agent.

        Args:
            agent_name: The name of the agent.
            behavior_id: Unique identifier for the behavior.
            behavior: The behavior description in "when...do..." format.

        Returns:
            True if successful, False otherwise.

        Raises:
            ValueError: If behavior format is invalid.
            IOError, TypeError, OSError: If reading or writing fails.
        """
        # Validate behavior format
        if not isinstance(behavior, str) or not behavior.strip():
            raise ValueError("Behavior must be a non-empty string")

        behavior_lower = behavior.lower().strip()
        if not behavior_lower.startswith("when"):
            raise ValueError("Behavior must follow 'when...do...' format")

        adaptive_data = self._read_json_file(
            self.adaptive_behaviors_file_path, default_value={}
        )

        if not isinstance(adaptive_data, dict):
            logger.warning(
                "WARNING: Adaptive behaviors file was not a dictionary. Resetting."
            )
            adaptive_data = {}

        # Initialize agent's behaviors if not exists
        if agent_name not in adaptive_data:
            adaptive_data[agent_name] = {}

        # Store the behavior
        adaptive_data[agent_name][behavior_id] = behavior.strip()

        try:
            self._write_json_file(self.adaptive_behaviors_file_path, adaptive_data)
            logger.info(
                f"INFO: Stored adaptive behavior '{behavior_id}' for agent '{agent_name}'"
            )
            return True
        except Exception as e:
            logger.error(f"ERROR: Failed to store adaptive behavior: {e}")
            return False

    def remove_adaptive_behavior(self, agent_name: str, behavior_id: str) -> bool:
        """
        Removes a specific adaptive behavior for an agent.

        Args:
            agent_name: The name of the agent.
            behavior_id: Unique identifier for the behavior to remove.

        Returns:
            True if successful or behavior didn't exist, False on error.
        """
        adaptive_data = self._read_json_file(
            self.adaptive_behaviors_file_path, default_value={}
        )

        if not isinstance(adaptive_data, dict):
            logger.warning("WARNING: Adaptive behaviors file was not a dictionary.")
            return True

        if agent_name in adaptive_data and behavior_id in adaptive_data[agent_name]:
            del adaptive_data[agent_name][behavior_id]

            # Clean up empty agent entries
            if not adaptive_data[agent_name]:
                del adaptive_data[agent_name]

            try:
                self._write_json_file(self.adaptive_behaviors_file_path, adaptive_data)
                logger.info(
                    f"INFO: Removed adaptive behavior '{behavior_id}' for agent '{agent_name}'"
                )
                return True
            except Exception as e:
                logger.error(f"ERROR: Failed to remove adaptive behavior: {e}")
                return False

        return True  # Behavior didn't exist, consider it successful

    def list_all_adaptive_behaviors(self) -> Dict[str, Dict[str, str]]:
        """
        Retrieves all adaptive behaviors for all agents.

        Returns:
            Dictionary mapping agent names to their behavior dictionaries.
        """
        adaptive_data = self._read_json_file(
            self.adaptive_behaviors_file_path, default_value={}
        )

        if not isinstance(adaptive_data, dict):
            logger.warning(
                "WARNING: Adaptive behaviors file was not a dictionary. Returning empty."
            )
            return {}

        return adaptive_data



from typing import Optional
from .config import MCPConfigManager
from AgentCrew.modules.mcpclient import MCPService
from AgentCrew.modules import logger


class MCPSessionManager:
    """Manager for MCP sessions and server connections."""

    _instance = None

    @classmethod
    def get_instance(cls):
        """Get the singleton instance of the session manager."""
        if cls._instance is None:
            cls._instance = MCPSessionManager()
        return cls._instance

    @classmethod
    def force_new_instance(cls):
        cls._instance = MCPSessionManager()
        return cls._instance

    def __init__(self):
        """Initialize the session manager."""
        self.config_manager = MCPConfigManager()
        self.mcp_service = MCPService()
        self.initialized = False

    def initialize(self) -> None:
        """Synchronous initialization method to start MCPService and initiate server connections."""
        if self.initialized:
            logger.info("MCPSessionManager: Already initialized.")
            return

        logger.info("MCPSessionManager: Initializing...")
        # Start the MCP service's event loop thread
        self.mcp_service.start()

        # Initialize server connections asynchronously on the mcp_service's loop
        # This call will block until initialize_servers_async is scheduled and returns.
        # The actual connections will happen in background tasks within MCPService.
        try:
            # self.mcp_service._run_async(self.initialize_servers_async())
            self.initialized = True
            logger.info("MCPSessionManager: Initialization process started.")
        except Exception as e:
            logger.error(f"MCPSessionManager: Error during async initialization: {e}")
            # Potentially stop mcp_service if its start was successful but async part failed
            self.mcp_service.stop()
            self.initialized = False  # Ensure it's marked as not initialized

    def initialize_for_agent(self, agent_name: Optional[str] = None) -> None:
        if not self.initialized:
            logger.error("MCPSessionManager: Has not initialized.")
            return

        try:
            self.mcp_service._run_async(
                self.mcp_service.shutdown_all_server_connections(agent_name)
            )
            self.mcp_service._run_async(self.initialize_servers_async(agent_name))
            logger.info("MCPSessionManager: Initialization process started.")
        except Exception as e:
            logger.error(f"MCPSessionManager: Error during async initialization: {e}")

    async def initialize_servers_async(self, agent_name: Optional[str] = None) -> None:
        """
        Asynchronously starts the connection management for all enabled MCP servers.
        This method is intended to be run on the MCPService's event loop.
        """
        # Load server configurations
        self.config_manager.load_config()  # Ensure configs are loaded
        enabled_servers = self.config_manager.get_enabled_servers(agent_name)

        if not enabled_servers:
            logger.info(
                "MCPSessionManager: No enabled MCP servers found in configuration."
            )
            return

        logger.info(
            f"MCPSessionManager: Found {len(enabled_servers)} enabled MCP servers. Requesting MCPService to manage connections..."
        )

        for server_id, config in enabled_servers.items():
            logger.info(
                f"MCPSessionManager: Requesting MCPService to manage connection for {server_id}"
            )
            # This is an async call. Since this method itself runs on the MCPService's loop (via _run_async),
            # we can directly await it.
            await self.mcp_service.start_server_connection_management(
                config, agent_name
            )

        logger.info(
            "MCPSessionManager: Finished requesting connection management for all enabled servers."
        )
        # Note: Actual connections are managed by background tasks in MCPService.

    def cleanup_for_agent(self, agent_name: str):
        if not self.initialized:
            logger.error("MCPSessionManager: Has not initialized.")
            return

        try:
            logger.info("MCPSessionManager: Starting cleanup...")
            self.mcp_service._run_async(
                self.mcp_service.shutdown_all_server_connections(agent_name)
            )
        except Exception as e:
            logger.error(
                f"MCPSessionManager: Error during cleanup_for_agent {agent_name}: {e}"
            )
        finally:
            logger.info("MCPSessionManager: Cleanup complete.")

    def cleanup(self):
        """Clean up all resources, including stopping MCP service and connections."""
        if not self.initialized and not (
            hasattr(self.mcp_service, "loop_thread")
            and self.mcp_service.loop_thread.is_alive()
        ):
            logger.info(
                "MCPSessionManager: Cleanup called but not initialized or service not running. Skipping."
            )
            return

        logger.info("MCPSessionManager: Starting cleanup...")
        try:
            # Signal all server connections to shut down and wait for them.
            # This needs to run on the MCPService's event loop.
            if hasattr(self.mcp_service, "loop") and self.mcp_service.loop.is_running():
                logger.info(
                    "MCPSessionManager: Running shutdown_all_server_connections on MCPService loop."
                )
                self.mcp_service._run_async(
                    self.mcp_service.shutdown_all_server_connections()
                )
            elif (
                hasattr(self.mcp_service, "loop_thread")
                and self.mcp_service.loop_thread.is_alive()
            ):
                logger.warning(
                    "MCPSessionManager: MCPService loop not running but thread alive. Attempting async shutdown."
                )
                # This case is tricky, _run_async might fail if loop isn't running.
                # However, shutdown_all_server_connections primarily sets events and gathers tasks
                # that might still be on a non-closed loop.
                try:
                    self.mcp_service._run_async(
                        self.mcp_service.shutdown_all_server_connections()
                    )
                except Exception as e_async_shutdown:
                    logger.error(
                        f"MCPSessionManager: Exception during _run_async for shutdown: {e_async_shutdown}"
                    )
            else:
                logger.warning(
                    "MCPSessionManager: MCPService loop not running/thread not alive, cannot run async shutdown. Resources might not be fully cleaned."
                )

            # Stop the MCPService's event loop and thread
            logger.info("MCPSessionManager: Stopping MCPService.")
            self.mcp_service.stop()  # This will join the thread

        except Exception:
            logger.exception("MCPSessionManager: Error during cleanup")
        finally:
            self.initialized = False
            logger.info("MCPSessionManager: Cleanup complete.")



from os import path as os_path
from typing import Dict, Any, Callable
from .service import CodeAnalysisService


# Tool definition function
def get_code_analysis_tool_definition(provider="claude") -> Dict[str, Any]:
    """
    Return the tool definition for code analysis based on provider.

    Args:
        provider: The LLM provider ("claude", "groq", or "openai")

    Returns:
        Dict containing the tool definition
    """
    description = "Analyzes the structure of source code files within a repository, creating a structural map. This identifies key code elements, enabling code understanding and project organization insights. Explain what insights you are hoping to gain from analyzing the repository before using this tool."

    tool_arguments = {
        "path": {
            "type": "string",
            "description": "The root directory to analyze. Use '.' to analyze all source files in the current directory, or specify a subdirectory (e.g., 'src') to analyze files within that directory. Choose the path that will provide the most relevant information for the task at hand.",
        }
    }
    tool_required = ["path"]

    if provider == "claude":
        return {
            "name": "analyze_repo",
            "description": description,
            "input_schema": {
                "type": "object",
                "properties": tool_arguments,
                "required": tool_required,
            },
        }
    else:  # provider == "openai"
        return {
            "type": "function",
            "function": {
                "name": "analyze_repo",
                "description": description,
                "parameters": {
                    "type": "object",
                    "properties": tool_arguments,
                    "required": tool_required,
                },
            },
        }


# Tool handler function
def get_code_analysis_tool_handler(
    code_analysis_service: CodeAnalysisService,
) -> Callable:
    """Return the handler function for the code analysis tool."""

    def handler(**params) -> str:
        path = params.get("path", ".")
        path = os_path.expanduser(path)
        result = code_analysis_service.analyze_code_structure(path)
        if isinstance(result, dict) and "error" in result:
            raise Exception(f"Failed to analyze code: {result['error']}")

        return result

    return handler


def get_file_content_tool_definition(provider="claude"):
    """
    Return the tool definition for retrieving file content based on provider.

    Args:
        provider: The LLM provider ("claude", "groq", or "openai")

    Returns:
        Dict containing the tool definition
    """
    tool_description = "Reads the content of a file, or a specific code element within that file (function or class body). Use this to examine the logic of specific functions, the structure of classes, or the overall content of a file."

    tool_arguments = {
        "file_path": {
            "type": "string",
            "description": "The relative path from the current directory of the agent to the local repository file. Example: 'src/my_module.py'",
        },
        "element_type": {
            "type": "string",
            "description": "The type of code element to extract. Use this when targeting a specific element.",
            "enum": ["class", "function"],
        },
        "element_name": {
            "type": "string",
            "description": "The name of the class or function to extract. Use this when targeting a specific element. Case-sensitive.",
        },
        "scope_path": {
            "type": "string",
            "description": "A dot-separated path to resolve ambiguity when multiple elements share the same name (e.g., 'ClassName.method_name'). Required only if the element name is ambiguous. Omit if unnecessary.",
        },
    }
    tool_required = ["file_path"]

    if provider == "claude":
        return {
            "name": "read_file",
            "description": tool_description,
            "input_schema": {
                "type": "object",
                "properties": tool_arguments,
                "required": tool_required,
            },
        }
    else:  # provider == "openai"
        return {
            "type": "function",
            "function": {
                "name": "read_file",
                "description": tool_description,
                "parameters": {
                    "type": "object",
                    "properties": tool_arguments,
                    "required": tool_required,
                },
            },
        }


def get_file_content_tool_handler(
    code_analysis_service: CodeAnalysisService,
):
    """Returns a function that handles the get_file_content tool."""

    def handler(**params) -> str:
        file_path = params.get("file_path")
        element_type = params.get("element_type")
        element_name = params.get("element_name")
        scope_path = params.get("scope_path")

        if not file_path:
            raise Exception("File path is required")

        # Validate parameters
        if element_type and element_type not in ["class", "function"]:
            raise Exception("Element type must be 'class' or 'function'")

        if (element_type and not element_name) or (element_name and not element_type):
            raise Exception(
                "Both element_type and element_name must be provided together"
            )

        results = code_analysis_service.get_file_content(
            file_path, element_type, element_name, scope_path
        )

        content = ""

        for path, code in results.items():
            content += f"{path}: {code}\n"

        # If we're getting a specific element, format the output accordingly
        if element_type and element_name:
            scope_info = f" in {scope_path}" if scope_path else ""
            return f"CONTENT OF {element_name} {element_type}{scope_info}: {content}"
        else:
            # If we're getting the whole file content
            return content

    return handler


def register(service_instance=None, agent=None):
    """
    Register this tool with the central registry or directly with an agent

    Args:
        service_instance: The code analysis service instance
        agent: Agent instance to register with directly (optional)
    """
    from AgentCrew.modules.tools.registration import register_tool

    register_tool(
        get_code_analysis_tool_definition,
        get_code_analysis_tool_handler,
        service_instance,
        agent,
    )

    register_tool(
        get_file_content_tool_definition,
        get_file_content_tool_handler,
        service_instance,
        agent,
    )



"""
Task management for A2A protocol implementation.
"""

import asyncio
from datetime import datetime
from typing import Dict, AsyncIterable, Optional, Any, Union
from AgentCrew.modules.agents import AgentManager, LocalAgent
from AgentCrew.modules.agents.base import MessageType
from AgentCrew.modules import logger
import tempfile
import os

from a2a.types import (
    CancelTaskResponse,
    GetTaskResponse,
    GetTaskSuccessResponse,
    JSONRPCError,
    JSONRPCErrorResponse,
    SendMessageResponse,
    SendStreamingMessageRequest,
    SendStreamingMessageResponse,
    SendStreamingMessageSuccessResponse,
    CancelTaskSuccessResponse,
    SetTaskPushNotificationConfigRequest,
    SetTaskPushNotificationConfigResponse,
    GetTaskPushNotificationConfigRequest,
    GetTaskPushNotificationConfigResponse,
    SendMessageSuccessResponse,
    TaskResubscriptionRequest,
    Task,
    TaskStatus,
    TaskState,
    SendMessageRequest,
    GetTaskRequest,
    CancelTaskRequest,
    JSONRPCResponse,
    TaskStatusUpdateEvent,
    TaskArtifactUpdateEvent,
    TaskNotFoundError,
    TaskNotCancelableError,
)
from .adapters import (
    convert_a2a_message_to_agent,
    convert_agent_response_to_a2a_artifact,
    convert_agent_message_to_a2a,
)
from .common.server.task_manager import TaskManager


class AgentTaskManager(TaskManager):
    """Manages tasks for a specific agent"""

    def __init__(self, agent_name: str, agent_manager: AgentManager):
        self.agent_name = agent_name
        self.agent_manager = agent_manager
        self.tasks: Dict[str, Task] = {}
        self.task_history: Dict[str, list[Dict[str, Any]]] = {}
        self.streaming_tasks: Dict[str, asyncio.Queue] = {}
        self.file_handler = None

        self.agent = self.agent_manager.get_agent(self.agent_name)
        if self.agent is None or not isinstance(self.agent, LocalAgent):
            raise ValueError(f"Agent {agent_name} not found or is not a LocalAgent")

        self.memory_service = self.agent.services["memory"]

    async def on_send_message(
        self, request: SendMessageRequest | SendStreamingMessageRequest
    ) -> SendMessageResponse:
        """
        Handle message/send request for this agent.

        Args:
            request: The message request

        Returns:
            JSON-RPC response with task result
        """
        if not self.agent or not isinstance(self.agent, LocalAgent):
            return SendMessageResponse(
                root=JSONRPCErrorResponse(
                    id=request.id,
                    error=JSONRPCError(
                        code=-32001, message=f"Agent {self.agent_name} not found"
                    ),
                )
            )

        # Generate task ID from message
        task_id = (
            request.params.message.task_id
            or f"task_{request.params.message.message_id}"
        )

        if task_id not in self.tasks:
            # Create task with initial state
            task = Task(
                id=task_id,
                context_id=request.params.message.context_id or f"ctx_{task_id}",
                status=TaskStatus(
                    state=TaskState.working, timestamp=datetime.now().isoformat()
                ),
            )
            self.tasks[task.id] = task

        task = self.tasks[task_id]
        if task_id not in self.task_history:
            self.task_history[task_id] = []

        # Convert A2A message to SwissKnife format
        message = convert_a2a_message_to_agent(request.params.message)
        if next(
            (m for m in message.get("content", []) if m.get("type", "text") == "file"),
            None,
        ):
            from AgentCrew.modules.chat.file_handler import FileHandler

            new_parts = []
            if self.file_handler is None:
                self.file_handler = FileHandler()
            for part in message.get("content", []):
                if part.get("type") == "file":
                    temp_file = os.path.join(tempfile.gettempdir(), part["file_name"])
                    with open(temp_file, "wb") as f:
                        f.write(part["file_data"])
                    new_parts.append(self.file_handler.process_file(temp_file))
                else:
                    new_parts.append(part)

            message["content"] = new_parts

        self.task_history[task_id].append(message)

        # Process with agent (non-blocking)
        asyncio.create_task(self._process_agent_task(self.agent, task))

        # Return initial task state
        return SendMessageResponse(
            root=SendMessageSuccessResponse(id=request.id, result=task)
        )

    async def on_send_message_streaming(
        self, request: SendStreamingMessageRequest
    ) -> Union[AsyncIterable[SendStreamingMessageResponse], JSONRPCResponse]:
        """
        Handle message/stream request for this agent.

        Args:
            request: The message request

        Yields:
            JSON-RPC responses with task updates
        """
        # Generate task ID from message
        task_id = (
            request.params.message.task_id
            or f"task_{request.params.message.message_id}"
        )

        # Create streaming queue
        queue = asyncio.Queue()
        self.streaming_tasks[task_id] = queue

        try:
            # Start the task
            response = await self.on_send_message(request)

            # If there was an error, yield it and stop
            if isinstance(response.root, JSONRPCErrorResponse):
                yield SendStreamingMessageResponse(root=response.root)
                return

            # Yield events from the queue
            while True:
                event = await queue.get()
                if event is None:  # End of stream
                    break
                yield SendStreamingMessageResponse(
                    root=SendStreamingMessageSuccessResponse(
                        id=request.id, result=event
                    )
                )

        finally:
            # Clean up
            self.streaming_tasks.pop(task_id, None)

    async def _process_agent_task(self, agent: LocalAgent, task: Task):
        """
        Process a task with the agent (background task).

        Args:
            agent: The agent to process the task
            message: The message to process
            task: The task object to update
        """
        try:
            artifacts = []
            if task.id not in self.task_history:
                raise ValueError("Task history is not existed")

            input_tokens = 0
            output_tokens = 0

            async def _process_task():
                # Process with agent

                # Create artifacts from response
                current_response = ""
                response_message = ""
                thinking_content = ""
                thinking_signature = ""
                tool_uses = []

                def process_result(_tool_uses, _input_tokens, _output_tokens):
                    nonlocal tool_uses, input_tokens, output_tokens
                    tool_uses = _tool_uses
                    input_tokens += _input_tokens
                    output_tokens += _output_tokens

                async for (
                    response_message,
                    chunk_text,
                    thinking_chunk,
                ) in agent.process_messages(
                    self.task_history[task.id], callback=process_result
                ):
                    # Update current response
                    if response_message:
                        current_response = response_message

                    # Update task status
                    task.status.state = TaskState.working
                    task.status.timestamp = datetime.now().isoformat()

                    # If this is a streaming task, send updates
                    if task.id in self.streaming_tasks:
                        queue = self.streaming_tasks[task.id]

                        # Send thinking update if available
                        if thinking_chunk:
                            think_text_chunk, signature = thinking_chunk
                            if think_text_chunk:
                                thinking_content += think_text_chunk
                                await queue.put(
                                    TaskStatusUpdateEvent(
                                        task_id=task.id,
                                        context_id=task.context_id,
                                        status=TaskStatus(
                                            state=TaskState.working,
                                            message=convert_agent_message_to_a2a(
                                                {
                                                    "role": "agent",
                                                    "content": think_text_chunk,
                                                },
                                                f"msg_thinking_{hash(think_text_chunk)}",
                                            ),
                                        ),
                                        final=False,
                                    )
                                )
                            if signature:
                                thinking_signature += signature

                        # Send chunk update
                        if chunk_text:
                            artifact = convert_agent_response_to_a2a_artifact(
                                chunk_text,
                                artifact_id=f"artifact_{task.id}_{len(artifacts)}",
                            )
                            await queue.put(
                                TaskArtifactUpdateEvent(
                                    task_id=task.id,
                                    context_id=task.context_id,
                                    artifact=artifact,
                                )
                            )

                if tool_uses and len(tool_uses) > 0:
                    if task.id in self.streaming_tasks:
                        queue = self.streaming_tasks[task.id]
                        artifact = convert_agent_response_to_a2a_artifact(
                            "",
                            artifact_id=f"artifact_{task.id}_{len(artifacts)}",
                            tool_uses=tool_uses,
                        )
                        await queue.put(
                            TaskArtifactUpdateEvent(
                                task_id=task.id,
                                context_id=task.context_id,
                                artifact=artifact,
                            )
                        )
                        # prevent the execute_tool_call take the control of event loop before queue has been process
                        await asyncio.sleep(0.7)

                    # Add thinking content as a separate message if available
                    thinking_data = (
                        (thinking_content, thinking_signature)
                        if thinking_content
                        else None
                    )
                    thinking_message = agent.format_message(
                        MessageType.Thinking, {"thinking": thinking_data}
                    )
                    if thinking_message:
                        self.task_history[task.id].append(thinking_message)

                    # Format assistant message with the response and tool uses
                    assistant_message = agent.format_message(
                        MessageType.Assistant,
                        {
                            "message": response_message,
                            "tool_uses": [
                                t for t in tool_uses if t["name"] != "transfer"
                            ],
                        },
                    )
                    if assistant_message:
                        self.task_history[task.id].append(assistant_message)

                    # Process each tool use
                    for tool_use in tool_uses:
                        try:
                            tool_result = await agent.execute_tool_call(
                                tool_use["name"],
                                tool_use["input"],
                            )

                            tool_result_message = agent.format_message(
                                MessageType.ToolResult,
                                {"tool_use": tool_use, "tool_result": tool_result},
                            )
                            if tool_result_message:
                                self.task_history[task.id].append(tool_result_message)

                        except Exception as e:
                            error_message = agent.format_message(
                                MessageType.ToolResult,
                                {
                                    "tool_use": tool_use,
                                    "tool_result": str(e),
                                    "is_error": True,
                                },
                            )
                            if error_message:
                                self.task_history[task.id].append(error_message)

                    return await _process_task()
                return current_response

            current_response = await _process_task()
            if current_response.strip():
                assistant_message = agent.format_message(
                    MessageType.Assistant,
                    {
                        "message": current_response,
                    },
                )
                if assistant_message:
                    self.task_history[task.id].append(assistant_message)
                user_message = (
                    self.task_history[task.id][0]
                    .get("content", [{}])[0]
                    .get("text", "")
                )
                self.memory_service.store_conversation(
                    user_message, current_response, self.agent_name
                )

            # Create artifact from final response
            artifact = convert_agent_response_to_a2a_artifact(
                current_response, artifact_id=f"artifact_{task.id}_final"
            )
            artifacts.append(artifact)

            # Update task with final state
            task.status.state = TaskState.completed
            task.status.timestamp = datetime.now().isoformat()
            task.artifacts = artifacts

            # If this is a streaming task, send final update
            if task.id in self.streaming_tasks:
                queue = self.streaming_tasks[task.id]

                # Send final status
                await queue.put(
                    TaskStatusUpdateEvent(
                        task_id=task.id,
                        context_id=task.context_id,
                        status=task.status,
                        final=True,
                    )
                )

                # Mark queue as done
                await queue.put(None)

        except Exception as e:
            logger.error(str(e))
            # Handle errors
            task.status.state = TaskState.failed
            task.status.timestamp = datetime.now().isoformat()

            # If this is a streaming task, send error
            if task.id in self.streaming_tasks:
                queue = self.streaming_tasks[task.id]
                await queue.put(
                    TaskStatusUpdateEvent(
                        task_id=task.id,
                        context_id=task.context_id,
                        status=task.status,
                        final=True,
                    )
                )
                await queue.put(None)

    async def on_get_task(self, request: GetTaskRequest) -> GetTaskResponse:
        """
        Handle tasks/get request for this agent.

        Args:
            request: The task request

        Returns:
            JSON-RPC response with task result
        """
        task_id = request.params.id
        if task_id not in self.tasks:
            return GetTaskResponse(
                root=JSONRPCErrorResponse(id=request.id, error=TaskNotFoundError())
            )

        return GetTaskResponse(
            root=GetTaskSuccessResponse(id=request.id, result=self.tasks[task_id])
        )

    async def on_cancel_task(self, request: CancelTaskRequest) -> CancelTaskResponse:
        """
        Handle tasks/cancel request for this agent.

        Args:
            request: The task request

        Returns:
            JSON-RPC response with task result
        """
        task_id = request.params.id
        if task_id not in self.tasks:
            return CancelTaskResponse(
                root=JSONRPCErrorResponse(id=request.id, error=TaskNotFoundError())
            )

        task = self.tasks[task_id]

        # Check if task can be canceled
        if task.status.state in [
            TaskState.completed,
            TaskState.failed,
            TaskState.canceled,
        ]:
            return CancelTaskResponse(
                root=JSONRPCErrorResponse(id=request.id, error=TaskNotCancelableError())
            )

        # Update task status
        task.status.state = TaskState.canceled
        task.status.timestamp = datetime.now().isoformat()

        # If this is a streaming task, send cancellation
        if task_id in self.streaming_tasks:
            queue = self.streaming_tasks[task_id]
            await queue.put(
                TaskStatusUpdateEvent(
                    task_id=task_id,
                    context_id=task.context_id,
                    status=task.status,
                    final=True,
                )
            )
            await queue.put(None)

        return CancelTaskResponse(
            root=CancelTaskSuccessResponse(id=request.id, result=task)
        )

    async def on_set_task_push_notification(
        self, request: SetTaskPushNotificationConfigRequest
    ) -> SetTaskPushNotificationConfigResponse:
        raise NotImplementedError("")

    async def on_get_task_push_notification(
        self, request: GetTaskPushNotificationConfigRequest
    ) -> GetTaskPushNotificationConfigResponse:
        raise NotImplementedError("")

    async def on_resubscribe_to_task(
        self, request: TaskResubscriptionRequest
    ) -> Union[AsyncIterable[SendStreamingMessageResponse], JSONRPCResponse]:
        raise NotImplementedError("")

    # Legacy methods for backward compatibility
    async def on_send_task(self, request: SendMessageRequest) -> SendMessageResponse:
        """Legacy method - delegates to on_send_message"""
        return await self.on_send_message(request)

    async def on_send_task_subscribe(
        self, request: SendStreamingMessageRequest
    ) -> Union[AsyncIterable[SendStreamingMessageResponse], JSONRPCResponse]:
        """Legacy method - delegates to on_send_message_streaming"""
        return await self.on_send_message_streaming(request)


class MultiAgentTaskManager:
    """Manages tasks for multiple agents"""

    def __init__(self, agent_manager: AgentManager):
        self.agent_manager = agent_manager
        self.agent_task_managers: Dict[str, AgentTaskManager] = {}

        # Initialize task managers for all agents
        for agent_name in agent_manager.agents:
            self.agent_task_managers[agent_name] = AgentTaskManager(
                agent_name, agent_manager
            )

    def get_task_manager(self, agent_name: str) -> Optional[AgentTaskManager]:
        """
        Get the task manager for a specific agent.

        Args:
            agent_name: Name of the agent

        Returns:
            The task manager if found, None otherwise
        """
        return self.agent_task_managers.get(agent_name)



from .transfer import (
    get_transfer_tool_definition,
    get_transfer_tool_handler,
    transfer_tool_prompt,
    register as register_transfer,
)

from .delegate import (
    get_delegate_tool_definition,
    get_delegate_tool_handler,
    delegate_tool_prompt,
    register as register_delegate,
)

__all__ = [
    "get_transfer_tool_definition",
    "get_transfer_tool_handler",
    "transfer_tool_prompt",
    "register_transfer",
    "get_delegate_tool_definition",
    "get_delegate_tool_handler",
    "register_delegate",
    "delegate_tool_prompt",
]



from PySide6.QtWidgets import QCompleter
from PySide6.QtCore import Qt
from AgentCrew.modules.llm.model_registry import ModelRegistry


class GuiModelCompleter:
    """GUI completer for model commands."""

    def __init__(self):
        self.registry = ModelRegistry.get_instance()

    def get_completions(self, text):
        """Get model completions for GUI."""
        if not text.startswith("/model "):
            return []

        word_after_command = text[7:]  # Remove "/model "

        # Get all available models from the registry
        all_models = []
        for provider in self.registry.get_providers():
            for model in self.registry.get_models_by_provider(provider):
                all_models.append((model.id, model.name, provider))

        # Filter models based on what the user has typed
        completions = []
        for model_id, model_name, provider in all_models:
            if model_id.startswith(word_after_command):
                completions.append(f"{provider}/{model_id}")

        return completions


class GuiAgentCompleter:
    """GUI completer for agent commands."""

    def __init__(self):
        from AgentCrew.modules.agents.manager import AgentManager

        self.agent_manager = AgentManager.get_instance()

    def get_completions(self, text):
        """Get agent completions for GUI."""
        if not text.startswith("/agent "):
            return []

        word_after_command = text[7:]  # Remove "/agent "

        # Get all available agents from the manager
        completions = []
        for agent_name, agent in self.agent_manager.agents.items():
            if agent_name.startswith(word_after_command):
                completions.append(agent_name)

        return completions


class GuiJumpCompleter:
    """GUI completer for jump commands."""

    def __init__(self, message_handler=None):
        self.message_handler = message_handler

    def get_completions(self, text):
        """Get jump completions for GUI."""
        if not text.startswith("/jump "):
            return []

        word_after_command = text[6:]  # Remove "/jump "

        conversation_turns = (
            self.message_handler.conversation_turns if self.message_handler else []
        )

        completions = []
        for i, turn in enumerate(conversation_turns, 1):
            turn_str = str(i)
            if turn_str.startswith(word_after_command):
                completions.append(turn_str)

        return completions


class GuiMCPCompleter:
    """GUI completer for MCP commands."""

    def __init__(self, message_handler=None):
        if message_handler:
            self.mcp_service = message_handler.mcp_manager.mcp_service
        else:
            self.mcp_service = None

    def get_completions(self, text):
        """Get MCP completions for GUI."""
        if not text.startswith("/mcp "):
            return []

        word_after_command = text[5:]  # Remove "/mcp "

        completions = []
        if self.mcp_service and hasattr(self.mcp_service, "server_prompts"):
            for server_id, prompts in self.mcp_service.server_prompts.items():
                for prompt in prompts:
                    prompt_name = getattr(prompt, "name", None) or prompt.get("name")
                    if prompt_name:
                        full_name = f"{server_id}/{prompt_name}"
                        if full_name.startswith(word_after_command):
                            completions.append(full_name)

        return completions


class GuiCommandCompleter:
    """GUI completer for all commands."""

    def __init__(self):
        self.commands = [
            "/clear",
            "/copy",
            "/debug",
            "/think",
            "/consolidate",
            "/unconsolidate",
            "/jump",
            "/agent",
            "/model",
            "/mcp",
            "/file",
            "/list",
            "/load",
            "/help",
            "/exit",
            "/quit",
        ]

    def get_completions(self, text):
        """Get command completions for GUI."""
        if not text.startswith("/"):
            return []

        completions = []
        for command in self.commands:
            if command.startswith(text):
                completions.append(command)

        return completions


class GuiChatCompleter:
    """Combined GUI completer for chat commands."""

    def __init__(self, message_handler=None):
        self.model_completer = GuiModelCompleter()
        self.agent_completer = GuiAgentCompleter()
        self.jump_completer = GuiJumpCompleter(message_handler)
        self.mcp_completer = GuiMCPCompleter(message_handler)
        self.command_completer = GuiCommandCompleter()

    def get_completions(self, text):
        """Get all completions for the given text."""
        if text.startswith("/model "):
            return self.model_completer.get_completions(text)
        elif text.startswith("/agent "):
            return self.agent_completer.get_completions(text)
        elif text.startswith("/jump "):
            return self.jump_completer.get_completions(text)
        elif text.startswith("/mcp "):
            return self.mcp_completer.get_completions(text)
        elif text.startswith("/"):
            return self.command_completer.get_completions(text)
        else:
            return []

    def create_qt_completer(self, parent=None):
        """Create a QCompleter instance for Qt widgets."""
        completer = QCompleter(parent)
        completer.setCompletionMode(QCompleter.CompletionMode.PopupCompletion)
        completer.setCaseSensitivity(Qt.CaseSensitivity.CaseSensitive)
        return completer
